/*+++++++++++++++++++++
.IDENTIFICATION sgml2.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    HTML, TeX
.KEYWORDS       LaTeX, HTML, HTTP, WWW
.VERSION  1.0   10-May-1998
.VERSION  1.1   20-May-1998: Allow regular expressions substitutions
.VERSION  1.2   29-Jun-1998: Added "ignore" as <DEF> option
.VERSION  1.3   18-Aug-1998: Use ${<} and ${>} to issue brackets.
.VERSION  1.4   28-Jun-1999: Transform \n innto blanks if not -keep_nl
.VERSION  1.5   02-Nov-1999: Try to upgrade the error message location
.VERSION  1.6   11-Sep-2000: Accept XML
.VERSION  1.7   20-Sep-2000: Always ignore comments <!--       -->
.VERSION  1.8   10-Feb-2001: Unquote the &amp &lt &gt &quot &apos in ${...}
.VERSION  1.9   04-Mar-2001: Transform ref= according to ID
.VERSION  2.0   07-Mar-2001: Can be used as a function.
.VERSION  2.1   10-Mar-2001: ${.sub} revisited: can accept buffer
			'g', etc. Added ${.fitsDISP}
.VERSION  2.2   14-Mar-2001: ${@name} = Attribute only (or NULL)
.VERSION  2.3   09-Nov-2001: -s(ilent) option
.VERSION  2.4   16-Nov-2001: ${.inject} -- ${$} (issue dollar)
.VERSION  2.5   04-Dec-2001: Verification non-clsed braces
.VERSION  2.6   14-May-2002: Added {.colspan}
.VERSION  2.7   06-Nov-2002: BUG transmitted by jfliu@astro.lsa.umich.edu
.VERSION  2.8   28-Jun-2003: ${%} ==> Edit a text; option -trim
		(bad recognition of ]]> in verbose mode) Flagged by V2.7
.VERSION  2.9   18-Feb-2004: -mime option (input contains a mime header)
.VERSION  2.91  26-Feb-2004: Accept attribute = "value"
.VERSION  3.0   11-May-2004: Accept attributes in dictionary
.VERSION  3.1   08-Sep-2004: Function sgml_set (set options)
.VERSION  3.2   04-Nov-2004: CDATA followed by a newline.
.VERSION  3.3   12-Dec-2005: getargs bug
.VERSION  3.4   21-Sep-2007: added ${.gobble}, changed fitsDISP
.VERSION  3.41  17-Jan-2008: Still an error in ]]> (end-of-verbatim)
.VERSION  3.5   21-Mar-2008: Accept <eos/> (instead of <eos />; ${.entity}
.VERSION  3.51  26-Jul-2008: DISP format is in lowercase for variable-length.
.VERSION  3.52  19-Aug-2008: (xml2fits.c) Child was not properly terminated...
.COMMENTS       This program converts an SGML text into ascii.
	It uses a stack of dictionaries containing definition of
	SGML tags (style <tag [args]...>, eventually terminated by </tag>)
	and sgml Entities (style &...;).

	The dictionary contains definitions like:
	&entity;=...translation...	[Case sensitivity]
	<tag>=...translation...		[Must be closed, case insensitive]
	<tag>:<synonym_tag> 		[Defines an alternate name]
	${symbol}=value			[May be within <tag>]
	${.action [arg]}#fct		[May be within <tag>]

	The translation includes plain text, and special arguments
	which are written ${  ...  } (actions). The basic actions are:
	${}              name of current tag, lowercase (e.g. a in <A href=...>
	${... [</etag>]} output the text between tag / end of tag
	${*} 	  Current tag (without the < > )
	${^} 	  be sure we're at a line beginning (issue newline otherwise)
	${<}	  Issue the <
	${>}	  Issue the >
	${$}	  Issue the $ -- ${$$} = issue the $$
	${\n}	  issue a newline
	${\ooo}	  issue character with provided octal code
	${!cmd}	  execute the Shell Command
	${@name}  issue attribute (or empty string) from current tag.
	${$name}  issue contents of saved variable (defined in {.save ...})
	${vname}  issue text corresponding to vname defined:
		  1. in current tag e.g. ${href} in <A href=...>
		  2. in a saved variable (previously defined in {.save ...}
		  3. in the dictionary
		  4. from the environment

	${.ignore} 	  	ignore input
	${.include file}  	insert a file
	${.inject var}  	insert a variable (buffer)
	${.reset}  		Set the line number to 1
	${.typein prompt} 	prompt user for a line
	${.typeout text} 	Type the text to stderr
	${.save [name] [</..>]}	saves into a named buffer
	${.clear [name]}  	empty a buffer
	${.kill [name]}	  	removes a named buffer
	${.append [name]} 	appends to a named buffer
	${.squeeze [name]} 	remove spaces and tabs into a blank
	${.trim [name]} 	remove leading/trailing spaces
	${.mark [name]}		Defines a marker
	${.exit [name]}		Be sure to exit all tags at mark

	${.upper}	Transforms to uppercase
	${.lower}	Transforms to lowercase
	${.basename text}	  Name of file (strip out directory path)
	${.entity text}	  	  Use &text; as entity
	${.notag}		  Ignore following tags.
	${.dirname text}	  Name of directory (strip out file name)
	${.bibjvp jnl vol page Let}  Edit as 14-bytes the Journal Volume Page
	${.fitsDISP}  		Gather datatype,width,precision e.g. F7.3
	${.colspan}  		Inserts (n-1) \t from colspan=n
	${.sub  /regexp/subs/ text}  Transforms according to a regular expr.
	${.gsub /regexp/subs/ text}  Global Transformation from regular expr.

	Basic tags which are known there:
	<DEF> ... </DEF> : Definition of new entities / tags, one per line
	<LIST>           : display the contents of dictionaries
	<save name=...> ... </save> : save as a named variable

How to replace comments by lines starting by # :

<!-->:${^}#${.gsub /\n/\n#/ ${comment}}

Summary of Functions:
1. BUFFER manipulations:
    BUF *buf_alloc(int size)
    void buf_clear(BUF *b)
    void buf_kill(BUF *b)
    void buf_zero(BUF *b)
    int  buf_free(BUF *b)		// Keep only the buffer stub
    int  buf_expand(BUF *b, int bytes)
    int  buf_append(BUF *b, char *text, int len)
    int  buf_achar(BUF *b, int c)
    int  buf_symbol(BUF *b, char *symbol) 	// Append Expandd Symbol
    BUF *var_find(char *name)			// Find a variable
    BUF *var_1find(char *name)			// true name = 1st word
    BUF *var_set(char *name, char *value)	// Create or use a variable
    BUF *var_append(char *name, char *value)	// Append to existing variable
    int  var_kill(char *name)
2. i/o from/to current file
    int  next_char()
    int  get_char()
    int  get_until(BUF *b, int stop_char)
    int  get_brace(BUF *b)
    int  PutChar(int c)
    int  PutString(char *s)
    int  PutText(char *s, int len)
3. Dictionaries manipulation
    void load_dic(char *list, int is_sorted)
    char *lookup(DIC *dic, char *s)
    DIC  *t1new(int items)
    void t1kill(DIC *dic)
4. Matching Functions
    int  display_input(char *end_tag)	// Interprets input until </endtag>
    int  match_symbol(char *symbol)	// e.g. ${...}
    int  match_entity()			// e.g. &...;
    int  match_tag()			// e.g. <A href=...>
    int  match_comment()		// e.g. <!-- ... -->
5. Callable Definitions
    PTR_FCT sgml_fct(char *symbol, PTR_FCT fct)	// Install user-supplied fct
    PTR_FCT sgml_comment(PTR_FCT fct)		// Associates comment
    char *sgml_attribute(char *ID, char *attr_name)
    int  sgml_conv(char *text)		// Apply Conversions from text

Help debugging definitions:
   alias pm "print input_mode, output_mode"
   alias pt "print current_tag"
   alias pc "print (char)c"
   alias po "print *collector"
   alias pp "print *parsed_buf"
   alias pi "print &inbuf->text[inbuf->pos]"
   alias pcom "print *comment_buf"

---------------*/

#define VERSION "3.51"
#ifndef ALLOW_IF
#define ALLOW_IF 0	/* Let know about ${.if's} */
#endif

#ifndef _ARGS
#ifdef __STDC__
#define _ARGS(A)      A       /* ANSI */
#else
#define _ARGS(A)      ()      /* Traditional */
#define const
#endif
#endif

#ifndef DEBUG
#define DEBUG 0
#endif

#if DEBUG
#define Return(v) { printf("=>%d\n", v) ; return(v) ; }
#else
#define Return(v) return(v)
#endif

#define BUFINCR	        512	/* MUST BE a POWER of 2	(increment)*/
#define MAX_FCT		 64	/* Maximal number of functions     */
#define LEN_TAG 	 80	/* Maximal length of variable name */
#define MAX_REG	  	 64	/* Maximal Number of Regular Expr. */

#ifdef  TRACE_MEMORY
#define malloc  osmalloc
#define free    osmfree
#define realloc osmexpand
#define strdup	osmdup
#else
#define  osmtrace()
#endif

#include <stdlib.h>
#include <string.h>	/* Standard string manipulations */
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>	/* File Control		*/
#include <math.h>
#ifndef O_BINARY	/* Unix machines only	*/
#include <unistd.h>
#define O_BINARY 0
#define POPEN    1	/* function exists...	*/
#define getch()		fgetc(stderr) /* getchar */
static char **latin2dos ;
#else
#include <dir.h>	/* MS-DOS Definitions	*/
#define	 wait(aint)	*(aint)=0
#define POPEN    0	/* no such function...	*/
static char *latin2dos[] = {
  "\260", "\260", ","   , ""   , "\""  , "..." , ""   , ""   , /* 128-135 */
  "^"   , "\260", "_"   , ""   , "OE"  , "\260", "\260", "\260", /* 136-143 */
  "\260", "`"   , "'"   , "\""  , "\""  , ""   , "" ,  ""   , /* 144-151 */
  "~"   , "TM"  , "_"   , ""   , "oe"  , "\260", "\260", "Y"   , /* 152-159 */
  "\377", ""   , ""   , ""   , ""   , ""   , "|"   , "s."  , /* 160-167 */
  "\""  , "(c)" , ""   , ""   , ""   , ""   , ".TM.", "-"   , /* 168-175 */
  ""   , ""   , ""   , "3"   , "'"   , ""   , "    ", ""   , /* 176-183 */
  ","   , "1"   , ""   , ""   , ""   , ""   , "3/4" , ""   , /* 184-191 */
  "A"   , "A"   , "A"   , "A"   , ""   , ""   , "AE"  , ""   , /* 192-199 */
  "E"   , ""   , "E"   , "E"   , "I"   , "I"   , "I"   , "I"   , /* 200-207 */
  "D"   , ""   , "O"   , "O"   , "O"   , "O"   , ""   , "x"   , /* 208-215 */
  "O"   , "U"   , "U"   , "U"   , ""   , "Y"   , "b"   , ""   , /* 216-223 */
  ""   , "a"   , ""   , ""   , ""   , ""   , "ae"  , ""   , /* 224-231 */
  ""   , ""   , ""   , ""   , ""   , ""   , ""   , ""   , /* 232-239 */
  "d"   , ""   , ""   , ""   , ""   , "o"   , ""   , ""   , /* 240-247 */
  "o"   , ""   , ""   , ""   , ""   , "y"   , "p"   , ""     /* 248-255 */
} ;
#endif

/* Accept 8-bit charactesr */
#define isGraph(c)	(c && (!isspace(c)))
static char tagChar[256];
#define isTagChar(c)	(tagChar[(c)&0xff])

static char oterm ;	/* for debugging getch	*/
static int lineno = 1 ;
static int colno = 0 ;

double atof() ;		/* Never know where this is declared....   */
double rint() ;		/* Never know where this is declared....   */
static char *lookup();
static char **lookup_entry, **lookup_last;
char *sgml_attribute() ;

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])
#define isid(c)		(isalnum(c)||((c)=='_'))
#define isquote(c)	(((c)=='\'')||((c)=='\"')||((c)=='`'))
#define issign(c)	(((c)=='+')||((c)=='-'))

#ifdef putchar
#undef putchar
#endif

typedef int (*SORT_FCT)(/* const void *, const void * */) ;
typedef char *PTR;
typedef PTR (*PTR_FCT)() ;	/* Function returning char pointer */
typedef int (*INT_FCT)() ;	/* Function returning integer      */

static PTR_FCT save_comment ;

/* The dictionaries are stacked */

typedef struct dictionary {
    struct dictionary *prev ;
    char *buf;
    int items, flags;
#define DIC_SORTED	1
#define DIC_MDEF	2	/* Accept multiple definitions	*/
#define DIC_IGNORE	0x80	/* Ignore undefined tags...	*/
    char *entries[2] ;		/* Actually items entries 	*/
} DIC ;
static DIC *top_dic ;
static char c1dic[] = "<&$#"; 	/* accepted as 1st char in dic	*/

/* 8-bit characters (Latin assumed) are kept as a plain array	*/
static char **char8b ;

/* The input is saved in a buffer */
typedef struct sbuf {
    int size, used, pos;	/* Allocated, used, current position 	*/
    char *text;			/* The actual size depends on size 	*/
} BUF;
static BUF  *collector;		/* When outputmode&2, keep the data	*/
static BUF  *inbuf;		/* When set, input saved here		*/
static BUF  *parsed_buf ;	/* For debuggin, saved what's read	*/
static BUF  *comment_buf ;	/* To collect the Comments		*/
static char *current_tag = "" ;	/* The current tag being examined	*/
static char *comment_tag = "<!-- \"comment\" -->" ;
static char *err_text ;		/* Text output in case of error	*/

static int colpos = 0 ;		/* Position in line, 0=BOL	*/
static int output_mode = 1 ;	/* Shared by output routines 	*/
static char *begline ;		/* Add at newlines in obyeyline	*/
#define MODE_buf	2	/* Results are collected	*/
#define MODE_verb	4	/* No interpretation (display)	*/
#define MODE_upper	0x10	/* Convert to uppercase		*/
#define MODE_lower	0x20	/* Convert to lowercase		*/
#define MODE_nl2cr	0x40	/* Gobble newlines (\n --> \r)	*/
#define MODE_obey	0x100	/* Obey to linemode		*/
#define MODE_notag	0x1000	/* No tag in input/output	*/
#define MODE_force	0x8000	/* Force the output		*/
#define MODE_nofile	0x8000	/* Input mode, no read ...	*/
#define MODE_rescan	1	/* Input mode, no read ...	*/
static int input_mode ;		/* Is 1 when called from rescan
					or match_comment 	*/

static char st_name[] = "(stdin)";
static FILE *infile;		/* Current Input File		*/
static char *thefile = st_name;	/*    "     "    FileName	*/
static char *theprogram;       	/*    "     "    Program 	*/
static FILE *_of ;		/* Current Output File		*/

/* Variables can be saved in Linked Buffers */
typedef struct saved_buf {
    struct saved_buf *prev ;
    char *name ;
    BUF buf ;
} NAMED_BUF ;
static NAMED_BUF *top_var ;
static int count_var;		/* Number of named buffers	*/

typedef struct marker {
    struct marker *prev ;
    char *name ;
    int value ;
    DIC *top_dic;
} MARKER ;
static MARKER *top_mark ;
static int display_depth ;
static int display_till ;	/* Limit display_input set by exec_exit */
static int display_stop ;	/* Set to -1 if display_input must stop */

/* Options to the Program */
static char osilent ;		/* -s option (silent)			*/
static char verbop = 0;
static char optnl = 0 ;		/* Keep new lines	*/
static char otrim = 0 ;		/* Remove leading/ending blanks		*/
static char omime = 0 ;		/* There is a mime header in the input	*/
static char otex = 0 ;		/* Translate into TeX (1) / softex (2)	*/
				/* Takes also value 8 = 8-bit Latin	*/
static int Argc; static char **Argv;

/* Definitions of 8-bit character equivalents */
static char *asc8[] = {
  "\200", "\201", ","   , "f"   , "\""  , "..." , "+"   , "++"  , /* 128-135 */
  "^"   , "\211", "_"   , "<"   , "OE"  , "\215", "\216", "\217", /* 136-143 */
  "\220", "`"   , "'"   , "\""  , "\""  , "o"   , "---" , "--"  , /* 144-151 */
  "~"   , "TM"  , "_"   , ">"   , "oe"  , "\235", "\236", "Y"   , /* 152-159 */
  " "   , "!"   , "ct"  , "L"   , "$"   , "Y"   , "|"   , "s."  , /* 160-167 */
  "\""  , "(c)",  "o"   , "\""  , "not.", "--"  , ".TM.", "-"   , /* 168-175 */
  "deg.", "+/-" , "2"   , "3"   , "'"   , "u"   , "    ", "."   , /* 176-183 */
  ","   , "1"   , "o"   , "\""  , "1/4" , "1/2" , "3/4" , "?"   , /* 184-191 */
  "A"   , "A"   , "A"   , "A"   , "Ae"  , "A"   , "AE"  , "C"   , /* 192-199 */
  "E"   , "E"   , "E"   , "E"   , "I"   , "I"   , "I"   , "I"   , /* 200-207 */
  "D"   , "N"   , "O"   , "O"   , "O"   , "O"   , "Oe"  , "x"   , /* 208-215 */
  "O"   , "U"   , "U"   , "U"   , "Ue"  , "Y"   , "b"   , "ss"  , /* 216-223 */
  "a"   , "a"   , "a"   , "a"   , "ae"  , "a"   , "ae"  , "c"   , /* 224-231 */
  "e"   , "e"   , "e"   , "e"   , "i"   , "i"   , "i"   , "i"   , /* 232-239 */
  "d"   , "n"   , "o"   , "o"   , "o"   , "o"   , "oe"  , ":"   , /* 240-247 */
  "o"   , "u"   , "u"   , "u"   , "ue"  , "y"   , "p"   , "y"     /* 248-255 */
} ;

/* Definitions of 8-bit character equivalents */
static char *tex8soft[] = {
  "\200", "\201", ",", "{\\em f}", "\"" , "...", "{dag}",
                                                        "{ddag}", /* 128-135 */
  "^", "{permil}", "_"  , "<"   , "OE"  , "\215", "\216", "\217", /* 136-143 */
  "\220", "`"   , "'"   , "``"  , "''"  , "{o}" , "--" ,   "-"  , /* 144-151 */
  "~"   , "TM"  , "_"   , ">"   , "oe"  , "\235", "\236", "Y"   , /* 152-159 */
  " "   , "!" , "{cent}", "{pounds}", "$" , "{Yen}", "|",
                                                   "{section}"  , /* 160-167 */
  "\""  , "{copyright}",  "^a^" , "``" , "{not}", "--"  , "{TM}", "-"   , /* 168-175 */
  "{deg}", "+/-", "^2^", "^3^", "'"   , "{mu}", "    ",   "."   , /* 176-183 */
  ","   , "^1^" , "^o^" , "''"  , "1/4" , "1/2" , "3/4" , "?"   , /* 184-191 */
  "\\`A", "\\'A", "\\^A", "\\~A", "\\\"A", "{AA}", "AE" , "C"   , /* 192-199 */
  "\\`E", "\\'E", "\\^E", "\\\"E",
                                 "\\`I", "\\'I", "\\^I", "\\\"I", /* 200-207 */
  "{Eth}", "\\~N", "\\`O", "\\'O", "\\^O", "\\~O", "\\\"O", "x" , /* 208-215 */
  "{Oslash}", "\\`U", "\\'U", "\\^U", "\\\"U",
                                         "\\'Y", "{Thorn}", "ss", /* 216-223 */
  "\\`a", "\\'a", "\\^a", "\\~a", "\\\"a",
                                          "{aa}", "ae", "\\c{c}", /* 224-231 */
  "\\`e", "\\'e", "\\^e", "\\\"e",
  		                 "\\`i", "\\'i", "\\^i", "\\\"i", /* 232-239 */
  "{eth}", "\\~n", "\\`o", "\\'o", "\\^o", "\\~o", "\\\"o", ":" , /* 240-247 */
  "{oslash}", "\\`u", "\\'u", "\\^u", "\\\"u",
                                   "\\'y", "{thorn}", "\\\"y"     /* 248-255 */
} ;

static char *tex8[] = {
  "\200", "\201", ","   , "{\\em f}", "''",
                              "{\\dots}", "{\\dag}", "{\\ddag}" , /* 128-135 */
  "\\^{ }", "\211", "_" , "<" , "{\\OE}", "\215", "\216", "\217", /* 136-143 */
  "\220", "`"   , "'"   , "``"  , "''", "{\bullet}", "---", "--", /* 144-151 */
  "\\~{ }", "^{TM}", "\\_", ">",
	                     "{\\oe}"  , "\235", "\236", "\"{Y}", /* 152-159 */
  "~"   , "!`", "{\\cents}", "{\\pounds}",
			       "\\$", "{\\Yens}", "|"  , "{\\S}", /* 160-167 */
  "\\\"{ }", "{\\copyright}.", "^a", "{\\langle\\!\\langle}",
                            "{\\not}", "--"  , "{\\TM}", "^-"   , /* 168-175 */
  "{\\deg}", "{\\pm}", "^2", "^3",
                         "\\'{ }", "{\\mu}", "{\\P}", "{\\cdot}", /* 176-183 */
  "\\c{ }", "^1", "^o"  , "{\\rangle\\!\\rangle}",
            "\\frac{1}{4}", "\\frac{1}{2}", "\\frac{3}{4}", "?`", /* 184-191 */
  "\\`{A}", "\\'{A}", "\\^{A}", "\\~{A}", "\\\"{A}", "\\{AA}",
                                           "{\\AE}", "\\c{C}"   , /* 192-199 */
  "\\`{E}", "\\'{E}", "\\^{E}", "\\\"{E}",
                         "\\`{I}", "\\'{I}", "\\^{I}", "\\\"{I}", /* 200-207 */
  "{\\Eth}", "\\~N", "\\`{O}", "\\'{O}",
                     "\\^{O}", "\\~{O}", "\\\"O", "{\\times}"   , /* 208-215 */
  "{\\O}", "\\`{U}", "\\'{U}", "\\^{U}", "\\\"U",
                               "\\'{Y}", "{\\Thorn}", "{\\ss}"  , /* 216-223 */
  "\\`{a}", "\\'{a}", "\\^{a}", "\\~{a}", "\\\"a",
                                    "{\\aa}", "{\\ae}", "\\c{c}", /* 224-231 */
  "\\`{e}", "\\'{e}", "\\^{e}", "\\\"{e}",
                 "\\`{\\i}", "\\'{\\i}", "\\^{\\i}", "\\\"{\\i}", /* 232-239 */
  "{\\eth}", "\\~{n}", "\\`{o}", "\\'{o}", "\\^{o}", "\\~{o}",
                                         "\\\'{o}", "{\\div}"   , /* 240-247 */
  "{\\o}", "\\`{u}", "\\'{u}", "\\^{u}", "\\\"u",
                               "\\'{y}", "{\\thorn}", "\\\"y"     /* 248-255 */
} ;


/* Definition of default tagfile (generates tex-like) :
   Remember that the separator can be
   = means a single scan of $ values
   : means multiple scans
   In the Action definition, the symbols mean
      / : compile the regular expression
      - : don't rescan the arguments.
*/
static char default_defs[] = "\
######### Definition of Actions\n\
${...}#12\n\
${.append}#15\n\
${.basename}#4\n\
${.bibjvp}#13\n\
${.clear}#17\n\
#${.collect}#9\n\
${.colspan}#29\n\
${.def}#-10\n\
${.delete}#16\n\
${.dirname}#5\n\
${.entity}#31\n\
${.exit}#21\n\
${.fitsDISP}#27\n\
${.gobble}#30\n\
${.gsub}#/-19\n\
${.ignore}#8\n\
${.include}#9\n\
${.inject}#28\n\
${.keep}#14\n\
${.kill}#16\n\
${.list}#11\n\
${.lower}#2\n\
${.mark}#20\n\
${.nl2cr}#6\n\
${.notag}#7\n\
${.obeyline}#24\n\
#${.popmode}#7\n\
#${.pushmode}#6\n\
${.reset}#23\n\
${.save}#14\n\
${.squeeze}#25\n\
${.sub}#/-18\n\
${.trim}#26\n\
${.typein}#3\n\
${.typeout}#22\n\
${.upper}#1\n\
######### Entities (case-sensitive ordering)\n\
&AElig;=\306\n\
&Aacute;=\301\n\
&Acirc;=\302\n\
&Agrave;=\300\n\
&Aring;=\305\n\
&Atilde;=\303\n\
&Auml;=\304\n\
&Ccedil;=\307\n\
&ETH;=\320\n\
&Eacute;=\311\n\
&Ecirc;=\312\n\
&Egrave;=\310\n\
&Euml;=\313\n\
&Iacute;=\315\n\
&Icirc;=\316\n\
&Igrave;=\314\n\
&Iuml;=\317\n\
&Ntilde;=\321\n\
&Oacute;=\323\n\
&Ocirc;=\324\n\
&Ograve;=\322\n\
&Oslash;=\330\n\
&Otilde;=\325\n\
&Ouml;=\326\n\
&THORN;=\336\n\
&Uacute;=\332\n\
&Ucirc;=\333\n\
&Ugrave;=\331\n\
&Uuml;=\334\n\
&Yacute;=\335\n\
&aacute;=\341\n\
&acirc;=\342\n\
&acute;=\264\n\
&aelig;=\346\n\
&agrave;=\340\n\
&amp;=&\n\
&apos;='\n\
&aring;=\345\n\
&atilde;=\343\n\
&auml;=\344\n\
&brvbar;=\246\n\
&bull;=\225\n\
&ccedil;=\347\n\
&cedil;=\270\n\
&cent;=\242\n\
&copy;=\251\n\
&cup3;=\263\n\
&curren;=\244\n\
&dagger;=\206\n\
&deg;=\260\n\
&divide;=\367\n\
&eacute;=\351\n\
&ecirc;=\352\n\
&egrave;=\350\n\
&eth;=\360\n\
&euml;=\353\n\
&frac12;=\275\n\
&frac14;=\274\n\
&frac34;=\276\n\
&gt;=>\n\
&hyphen;=-\n\
&iacute;=\355\n\
&icirc;=\356\n\
&iexcl;=\241\n\
&igrave;=\354\n\
&iquest;=\277\n\
&iuml;=\357\n\
&laquo;=\253\n\
&ldquo;=\223\n\
&lsquo;=\221\n\
&lt;=<\n\
&macr;=\257\n\
&mdash;=\226\n\
&micro;=\265\n\
&middot;=\267\n\
&minus;=-\n\
&nbsp;=\240\n\
&ndash;=\227\n\
&not;=\254\n\
&ntilde;=\361\n\
&oacute;=\363\n\
&ocirc;=\364\n\
&ograve;=\362\n\
&ordf;=\252\n\
&ordm;=\272\n\
&oslash;=\370\n\
&otilde;=\365\n\
&ouml;=\366\n\
&para;=\266\n\
&permil;=\211\n\
&plus;=+\n\
&plusmn;=\261\n\
&pound;=\243\n\
&quot;=\"\n\
&raquo;=\273\n\
&rdquo;=\224\n\
&reg;=\256\n\
&rsquo;=\222\n\
&sect;=\247\n\
&shy;=\255\n\
&sup1;=\271\n\
&sup2;=\262\n\
&szlig;=\337\n\
&thorn;=\376\n\
&tilde;=\230\n\
&times;=\327\n\
&uacute;=\372\n\
&ucirc;=\373\n\
&ugrave;=\371\n\
&uml;=\250\n\
&uuml;=\374\n\
&yacute;=\375\n\
&yen;=\245\n\
&yuml;=\377\n\
######### Default ELEMENT (tag) definitions (case insensitive)\n\
<!-->:\n\
<BR>:${\\n}\n\
<DEF>=${.def ${...}}\n\
<HR>:${^}--------------------------------------------------------------------\
--------${\\n}\n\
<LIST>:${.list Full List of Known Dictionaries:}\n\
<P>:${\\n}\n\
<PRE>=${.obeyline}${...}\n\
<SAVE>=${.save ${name}}${...}\n\
" ;

static char tex2_tags[] = "\
######### Definitions for tex2\n\
&ap;=~\n\
&copy;={copyright}\n\
&ga;={>~}\n\
&ge;={>=}\n\
&gsim;={>~}\n\
&la;={<~}\n\
&le;={<=}\n\
&lsim;={<~}\n\
&minus;=-\n\
&para;={P}\n\
&prop;={prop}\n\
&rsim;={>~}\n\
&sect;={S}\n\
&sim;=~\n\
<A>=${href}(${...})\n\
<B>={\\bf ${...}}\n\
<BR>:${\\n}\n\
<EM>={\\em ${...}}\n\
<HR>:{\\hr}\n\
<I>={\\it ${...}}\n\
<IMG>:\\tag{${*}}\n\
<LISTING>=\\begin{verbatim}${...}\\end{verbatim}\n\
<META>=\\tag{${*}}\n\
<P>:{\\par}\n\
<SUB>=_${...}_\n\
<SUP>=^${...}^\n\
<TABLE>=${...}${^}\n\
<TD>=${.colspan}${...}\t\n\
<TH>=${...}\t\n\
<TR>=${^}${...}\n\
<TT>={\\tt ${...}}\n\
<U>=\\underline{${...}}\n\
########## End of Definitions ######\n\
";

static char asc_tags[] =  "\n\
######### Definitions for ascii\n\
&ap;=~\n\
&ga;=>~\n\
&ge;=>=\n\
&la;=<~\n\
&le;=<=\n\
&lsim;=<~\n\
&minus;=-\n\
&prop;={prop}\n\
&rsim;=>~\n\
&sim;=~\n\
<A>=${href}(${...})\n\
<B>=${...}\n\
<BR>:${\\n}\n\
<EM>=${...}\n\
<FONT>=${...}\n\
<I>=${...}\n\
<P>:${^}${\\n}\n\
<SUB>=_${...}_\n\
<SUP>=^${...}^\n\
<TABLE>=${...}${^}\n\
<TD>=${.colspan}${...}\t\n\
<TH>=${...}\t\n\
<TR>=${^}${...}\n\
<TT>=${...}\n\
########## End of Definitions ######\n\
";

static char tex_tags[] = "\
######### Definitions for tex\n\
&amp;=\\&\n\
&ap;={\\approx}\n\
&ge;=\\ge \n\
&le;=\\le \n\
&lsim;=\\la \n\
&minus;=$-$\n\
&ndash;=$-$\n\
&prop;=\\propto \n\
&rsim;=\\ga \n\
<A>=\\A{${href}}{${...}}\n\
<B>={\\bf ${...}}\n\
<EM>={\\em ${...}}\n\
<HR>={\\hr}\n\
<I>={\\it ${...}}\n\
<IMG>=\\tag{${*}}\n\
<LI>=\\item \n\
<LISTING>=\\begin{verbatim}${...}\\end{verbatim}\n\
<META>=\\tag{${*}}\n\
<OL>=\\begin{enumerate}${...}\\end{enumerate}\n\
<P>={\\par}\n\
<SUB>=_{${...}}\n\
<SUP>=^{${...}}\n\
<TABLE>=\\begin{tabular}{|ccccc|}\\hline${\\n}${...}${^}\\hline\\end{tabular}\n\
<TD>=${.colspan p}${...}${.colspan /}&\n\
<TH>=\\multicolumn{1}{c}{\\bf${...}}&\n\
<TR>=${...}\\\\ \n\
<TT>={\\tt ${...}}\n\
<U>=\\underline{${...}}\n\
<UL>=\\begin{itemize}${...}\\end{itemize}\n\
######################## End of Definitions ######\n\
";

static char alatex_tag[]  = "<TAG>=\\begin{${}}${...}\\end{${}}" ;
static char alasoft_tag[] = "<TAG>={\\${} ${...}}" ;

static int indent = 0 ;
static int owidth = 0 ;	/* -w option */

static char usage[] = "\
Usage: sgml2 [-v[v]] [-s] [-help] [-list] [-a8|-tex|-tex2] [-trim] [-mime]\n\
       [-f[s|i] def_file] [-keep_nl] [-w w,i] [-e text] [file...] [-]" ;
static char help[] = "\
This program converts SGML into LaTeX (-tex) or plain ascii (or tsv).\n\
User-definitions of SGML tags can be gathered into a dictionary, specified by\n\
the -f option (the dictionary is assumed to be sorted list with -fs option, \n\
or ignore all non-existing tags with -fi option),\n\
or within the <DEF [sorted] [ignore]> ... </DEF> tags.\n\
   A dictionary consists of definitions of entities (&...;) or tags (<...>)\n\
one line per definition. The equivalence is preceded by = or :\n\
The contents of dictionaries can be displayed at any time with the <LIST> tag\n\
   The -trim option asks to squeeze the blanks in the XML tags.\n\
   The -mime option tells a mime header in included, is jsut passed over\n\
   The -w option asks to reformat the output, as lines of width w, indented\n\
by i bytes. However, the actions ${^} and ${\\n} do not add the indentation\n\
   The -e option defines the text to output where there is an error found.\n\
   The -s option asks for silent output.\n\
";

/* There is a need to have a special function which provides the
   next argument
*/

/* Declaration of recursive functions */
static BUF  *var_find() ;
static char *display_input() ;
static char *def_tag() ;
static char *rescan() ;
static int   match_symbol() ;
#if ALLOW_IF
static char *match_if() ;
#endif

/*==================================================================
		Very basic utilities...
 *==================================================================*/

static char *i2a(int val)
/*++++++++++++++++
.PURPOSE  Edit (in local buffer) the number
.RETURNS  Edited number
-----------------*/
{
  static char anum[20] ;
    sprintf(anum, "%d", val);
    return(anum) ;
}

static double a2f(char *text)
/*++++++++++++++++
.PURPOSE  Interpret a text for a double
.RETURNS  The value
-----------------*/
{
  double v, atof();
  char *p;
    for (p=text; isspace(*p); p++) ;
    v = atof(p);
    if ((*p == '-') || (*p == '+')) p++;
    while (isdigit(*p)) p++;
    if (*p == '.') for (++p; isdigit(*p); p++) ;
    if (tolower(*p) == 'e') {
	if ((*p == '-') || (*p == '+')) p++;
	while (isdigit(*p)) p++;
    }
    while (isspace(*p)) p++;
    if (*p) {
	fprintf(stderr, "++++a2f error: ");
	while (text < p) fputc(*text, stderr);
	fprintf(stderr, "`%c'%s\n", *p, p+1);
    }
    return(v);
}

static int strloc(char *text, int c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
{
  char *s;
    for ( s = text; *s; s++)      if (*s == c)    break;
    return(s-text);
}

static int stundiff(char *s1, char *s2, int len)
/*++++++++++++++++
.PURPOSE  Compare (case-insensitive) 2 strings
.RETURNS  Difference
-----------------*/
{
  int diff ; char *e1, *e2 ;
    for (e1 = s1+len, e2 = s2+len; (s1<e1) && (s2<e2); s1++, s2++) {
    	diff = toupper(*s1) - toupper(*s2) ;
    	if (diff) return(diff) ;
    }
    return(0) ;
}

static int thebyte(char *ins)
/*++++++++++++++++
.PURPOSE  Interpret a string according to C standards
.RETURNS  Value of byte
.REMARKS  \octal or \f \b \r \n ...
-----------------*/
{
  int o;
    if (*ins != '\\') return(*ins);
    switch(ins[1]) {
      case 'b': return('\b');
      case 'f': return('\f');
      case 'n': return('\n');
      case 'r': return('\r');
      case 't': return('\t');
      case 'e': case 'E': return(033);
      default:
	if (isdigit(ins[1]))
	    for (o=0, ++ins; isdigit(*ins); ins++) o = o*8 + (*ins-'0');
	else o = ins[1];
	return(o);
    }
}

static int strquote(char *s)
/*++++++++++++++++
.PURPOSE  Find the matching quote
.RETURNS  Offset of next "
-----------------*/
{
  char *p, q;
    p = s; q = *(p++) ;
    p += strloc(p, q) ;
    return(p-s) ;
}

static int strangle(char *text)
/*++++++++++++++++
.PURPOSE  Locate the closing > (the opening < is skipped)
.RETURNS  Index of  closing brace
-----------------*/
{
  int depth; char *p ;
    for (depth = 1, p = text; depth && *p ; p++) switch(*p) {
    	case '<': depth++; continue ;
    	case '>': depth-- ; continue ;
    	case '\'': case '"': case '`':	/* Quote */
    	  p += strquote(p) ;
    	  if (! *p) p--;
    	  continue ;
    }
    if (depth == 0) p-- ;
    return(p-text);
}

static int strbrace(char *text)
/*++++++++++++++++
.PURPOSE  Locate the closing brace (the opening brace is skipped)
.RETURNS  Index of  closing brace
-----------------*/
{
  int depth; char *p ;
    for (depth = 1, p = text; depth && *p ; p++) switch(*p) {
    	case '{': depth++; continue ;
    	case '}': depth-- ; continue ;
    	case '\'': case '"': case '`':	/* Quote */
    	  p += strquote(p) ;
    	  if (! *p) p--;
    	  continue ;
    }
    if (depth == 0) p-- ;
    return(p-text);
}

static int str2int(char *text, int *value)
/*++++++++++++++++
.PURPOSE  Extract a number from text
.RETURNS  Number of bytes scanned
-----------------*/
{
  char *s;
    for (s = text, *value=0; isdigit(*s); s++)
	*value = (*value * 10) + (*s - '0');
    return(s-text);
}

static int wordlen(char *s)
/*++++++++++++++++
.PURPOSE  According to first byte, determine length of word.
.RETURNS  The length
.REMARKS  ".." ==> quotes included, ${...}, or just word.
-----------------*/
{
  char *p ; int len ;
    if (isspace(*s)) return(0) ;	/* Nothing works... */
    /* Is it a quoted word ? 	*/
    if (isquote(*s)) {
        len = strquote(s) ;
        if (s[len]) len++ ;
        return(len) ;
    }
    /* ... or a named variable?	*/
    if ((s[0] == '$') && (s[1] == '{' /*}*/)) {
    	len = 2 + strquote(s+2) ;
    	if (s[len]) len++ ;
    	return(len) ;
    }
    /* ... take the word...	*/
    for (p=s; isGraph(*p); p++) ;
    return(p-s) ;
}

static char **getargs(char *s, int max_arg, int opt)
/*++++++++++++++++
.PURPOSE  Get the set of Arguments.
.RETURNS  An array to arguments of newly allocated pieces of memory.
.REMARKS  ".." ==> quotes included, ${...}, or just word.
	The last argument takes all remaining bytes.
	opt is set to 1 if rescan is asked for.
-----------------*/
{
  static char *argv[20];
  char *p, *a ; int n, len ;

    if (verbop&2) printf("....getargs(%p='%s', %d, %d)", s, 
       s ? s : "", max_arg, opt);
    if (max_arg >= ITEMS(argv)) {
	fprintf(stderr, "****getargs: limited to %d arguments (was %d)\n",
	    ITEMS(argv)-1, max_arg) ;
	max_arg = ITEMS(argv)-1 ;
    }
    if (max_arg <= 0) {
	fprintf(stderr, "****getargs: ask %d arguments from '%s'\n",
	    max_arg, s) ;
	return((char **)0) ;
    }
    n = 0 ;
    if (!s) s = "" ;
    p = s ;
    --max_arg ;

    /* Skip spaces. For non-last arguments, take word or quote */
    while (isspace(*s)) s++ ;
    while ((n <= max_arg) && (*s)) {
	p = s ;
	if ((n == max_arg) && (!isquote(*s)))
	     len = strlen(s) ;
	else len = wordlen(s) ;
	s += len ;
	if (isquote(*p)) p++, n -= 2 ;
	strncpy(a = argv[n] = malloc(len+1), p, len) ;
	a[len] = 0 ;
	if (opt && strchr(a, '$')) {
	    argv[n] = rescan(a, -1) ;
	    free(a) ;
	}
	while (isspace(*s)) s++ ;
	n++ ;
    }
    max_arg++ ;
    while (n <= max_arg) argv[n++] = (char *)0 ;
    if (*s) fprintf(stderr,
	"++++getargs: ignored trailing argument ...%s\n", s) ;

    if (verbop&2) {
	printf("\n        => ");
	for (n=0; n <= max_arg; n++) printf(" %p", argv[n]);
	printf("\n");
    }
    return(argv) ;
}

static char *trstr(char *str, int k) 	/* lower / upper / nl2cr */
/*++++++++++++++++
.PURPOSE  Trasnforms a string into uppercase / lowercase
.RETURNS  Translated string
.REMARKS  Newly allocated
-----------------*/
{
  static char *as = (char *)0; static int ls = 0 ;
  char *a, *s ; int len ;
    if (!str) {		/* No data. Just set up the flag */
    	/*
	output_mode &= ~(MODE_upper|MODE_lower) ;
    	if (k) output_mode |= MODE_upper ;
    	else   output_mode |= MODE_lower ;
	*/
	output_mode |= k;
    	return("") ;
    }
    len = strlen(str) + 1 ;
    if (len > ls) {
	ls = ((len+15)>>4)<<4 ;		/* Multiple of 16 bytes */
	if (!as) as = malloc(ls) ;
	else as = realloc(as, ls) ;
    }
    for (a=as, s=str ; *s; a++, s++) {
	len = *s&0xff;
	if (k&MODE_nl2cr) { if (len == '\n') len = '\r'; }
	if (k&MODE_upper)      *a = toupper(len);
	else if (k&MODE_lower) *a = tolower(len);
	else *a = len;
    }
    *a = 0 ;
    return(as) ;
}

static int strarg(char *s, char *name)
/*++++++++++++++++
.PURPOSE  Find location in input string s, of "NAME=value"
.RETURNS  Offset of name=
.REMARKS  name is case insensitive. Terminated by }
-----------------*/
{
  char *p ; int c, len ;
    /* Name excludes spaces, { } , quotes */
    for (p=name; isGraph(*p) && (!isquote(*p)) && (*p != '{') && (*p != '}');
       p++) ;
    len = p - name ;

    p = s;
    while(*p) {
    	/* Skip the current definition */
    	while (isGraph(*p)) {
    	    if (isquote(*p)) p += strquote(p) ;
    	    if (*p) p++ ;
    	}
    	while (isspace(*p)) p++ ;
    	if (stundiff(p, name, len) == 0) {
	    if (!(c=p[len])) break ;
	    if (c == '=') break ;
	    if (c == '>') break ;
	    if (isspace(c)) break ;
	}
    }
    /* printf("....strarg returns: %d\n", p-s) ; */
    return(p-s) ;
}

#if ALLOW_IF
/*==================================================================
		Pattern Matching (filenames rules)
 *==================================================================*/

int strmatch(char *template, char *name)
/*++++++++++++++++
.PURPOSE  Check name against (file) template
.RETURNS  1 (Match) / 0 (mismatch)
.REMARKS  * ? [...] recognized as wild chars
-----------------*/
{
  char *t, *f; int m;
    for (t = template, f = name; *t; t++) switch(*t) {
      case '?':			/* Matches one character */
	if (*f) f++;
	else return(0);
	continue;
      case '*':			/* Match 0 or n characters */
	if (! *++t) return(1);	/* The * is the last char of template */
#if 1		/* Try LONGEST possible matching */
	for (m = strlen(f); --m >= 0; ) {
	    if (strmatch(t, f+m))
		return(1);
	}
#else		/* Try SHORTEST possible matching */
	for (; *f; f++) {	/* Look for first matching char */
	    if (isalnum(*t) && (*f != *t)) continue;
	    if (strmatch(t,f))	/* Recursive call */
		return(1);
	}
#endif
	return(0);
      case '[':		/* A set of characters is provided */
	++t; m=0;
	if (*t == ']') { /* Closing brace is acceptable	*/
	    t++;
	    m = (*f == ']');
	}
	while ((m==0) && *t) {	/* Check: does *f be in set ? 	*/
	    if (*t == ']') return(0);		/* End of set	*/
	    if ((t[1] == '-') && (t[2] != ']')) {      /* Range	*/
		m = (*f >= t[0]) && (*f <= t[2]);
		t += 3;
	    }
	    else m = (*f == *t), t++;
	}
	if (m == 0) return(0);
	t += strloc(t, ']');
	f++;
	continue;
      case '\\':	/* Escaped Char. NO BREAK HERE! */
	t++;
      default:
	if (*t != *f) return(0);
	f++;
	continue;
    }
    return((!*t) && (!*f));	/* Returns 1 if *t AND *f are NUL */
}
#endif

/*==================================================================
		Regular Expressions
 *==================================================================*/

static char *pregexp ;
#define INIT
/* #define INIT		register char *pregexp = instring ; */
#define GETC()		*(pregexp++)
#define PEEKC()		(*pregexp)
#define UNGETC(c)	(--pregexp)
#define RETURN(c)	return(c);
#define	ERROR(c)	fprintf(_of, "\n\\verb\177**Error #%d in regexp '%s'+%d \177\n",\
			c,instring,(pregexp-instring))
#include <regexp.h>
#if 0
static char *oreg[MAX_REG] ;	/* Non-compiled regular expressions	*/
static char *areg[MAX_REG] ;	/* Array of Saved Complied Reg. Expr.	*/
static int  nreg ;		/* Number of elements in areg and oreg	*/
#else
static char compiled_reg[1024];	/* Compiled regular expression 		*/
#endif

int strexp(char *cexp, char *input)
/*++++++++++++++++
.PURPOSE  Check whether input satisfies the compiled regular expression
.RETURNS  Number of matched chars / 0 for mismatch
-----------------*/
{
  int saved_circ, n;
    saved_circ = circf ;
    circf = 1 ;
    if (step(input, cexp)) n = (loc2-loc1);
    else n = 0;
    circf = saved_circ ;
    return(n);
}

static int len_expr(char *input)
/*++++++++++++++++
.PURPOSE  Find the length of a regular expression
.RETURNS  The index of / in input
.REMARKS  input DOES NOT start by sep
-----------------*/
{
  char *p ;
  int in_brackets ;
    for (p=input, in_brackets=0; *p; p++) {
	if (*p == '/') {
	    if (in_brackets == 0) break ;
	    continue ;
	}
	if (*p == '[') {
	    in_brackets = 1 ;
	    if (p[1]) p++; 	/* To match []] */
	    continue ;
	}
	if (*p == ']') in_brackets = 0 ;
	else if (*p == '\\') {
	    if (p[1]) p++; 	/* To match \/  */
	}
    }
    return(p-input) ;
}

#if 0
static int find_expr(char *input)
/*++++++++++++++++
.PURPOSE  Fidn a regular expression, if it exists
.RETURNS  The index of the compiled expression in areg
.REMARKS  The separator must be a /  -- input starts by /
-----------------*/
{
  int n, len;
    len = len_expr(input) ;
    for (n=0; n<nreg; n++) {
	if (strnmatch(input, oreg[n], len) != 0) continue ;
	if (input[len] == '/') return(n) ;
    }
    return(-1) ;
}

static int compile_expr(char *input)
/*++++++++++++++++
.PURPOSE  Compile a regular expression
.RETURNS  How many bytes are parsed --- normally index of /
.REMARKS  The regular expression's address is saved in areg[nreg-1]
-----------------*/
{
  char buf[BUFSIZ] ;
  char *p, *e; int n;
    if (verbop) printf("----Compile '%s', nreg=%d\n", input, nreg) ;
    if (nreg >= MAX_REG) {
        fprintf(stderr, "****Too many (>%d) regular expressions ...%s\n",
            nreg, input) ;
        return(-1) ;
    }
    if (!*input) {	/* Invalid Regular Expression... */
        fprintf(stderr, "****Empty Regular Expression\n") ;
        return(-1) ;
    }
    memset(buf, 0, sizeof(buf));
    pregexp=input+1 ;	/* Keep the position in regexp	*/
    e = compile(pregexp, buf, buf+sizeof(buf), *input);
    n = 1 + (e-buf) ;	/* Keep 1 byte more for compiled regexp */
    p = malloc(n + 1 + (pregexp-input));
    memcpy(p, buf, n);
    areg[nreg] = p ;
    /* Save the original regular expression */
    p += n ;
    n = pregexp-input ;
    oreg[nreg] = p ;
    strncpy(p, input, n) ;
    return(n) ;
}
#else
static int compile_expr(char *input)
/*++++++++++++++++
.PURPOSE  Compile a regular expression (into compiled_reg)
.RETURNS  How many bytes are parsed --- normally index of /
-----------------*/
{
  char *e; 
    if (verbop) printf("----Compile '%s'\n", input) ;
    if (!*input) {	/* Invalid Regular Expression... */
        fprintf(stderr, "****Empty Regular Expression\n") ;
        return(-1) ;
    }
    memset(compiled_reg, 0, sizeof(compiled_reg));
    pregexp=input+1 ;	/* Keep the position in regexp	*/
    e = compile(pregexp, compiled_reg, compiled_reg+sizeof(compiled_reg),
	*input);
    return(pregexp-input) ;
}
#endif

/*==================================================================
		File operations
 *==================================================================*/
static char *inloc()
/*++++++++++++++++
.PURPOSE  Defines the position in input file
.RETURNS  Text
-----------------*/
{
  static char fmt[] = "%d+%d" ;
  static char buf[256] ;
    buf[0] = 0;
    if (thefile && (!err_text))
	strncpy(buf, thefile, sizeof(buf)-sizeof(fmt)-20) ;
    buf[sizeof(buf)-sizeof(fmt)-20] = 0;
    sprintf(buf+strlen(buf), fmt, lineno, colno) ;
    return(buf) ;
}

static char *load_file(char *name, int show)
/*++++++++++++++++
.PURPOSE  Load a file completely into memory
.RETURNS  The loaded text terminated by NULL / NULL if error
-----------------*/
{
  int  i;
  long count;
  char *buffer;

	/* Load file into Memory */
    i = open(name, O_BINARY);
    if (i <= 0) {
    	if (show)  {
	    perror(name),
	    fprintf(_of, "\n\\verb\177**Can't load file: %s\177\n", name) ;
	}
    	return ((char *)0);
    }
    count = lseek(i, 0L, 2); lseek(i, 0L, 0);

    if ((buffer = malloc(count+1))) {
    	if (read(i, buffer, count) != count) {
    	    perror(name);
    	    free(buffer);
    	    return ((char *)0);
    	}
    	buffer[count] = 0;
    }
    close(i);
    return(buffer);
}

static char *include_file(char *name)
/*++++++++++++++++
.PURPOSE  Execute the ${.include}
.RETURNS  ""
.REMARKS  Call display_input from file
-----------------*/
{
  BUF *old_inbuf ;
  FILE *old_input, *f ;
  char *old_name ;
  int old_lineno, old_colno, old_mode ;

    /* Try to open file file... */
    f = fopen(name, "r") ;
    if (!f) {
	fprintf(stderr, "\n****") ;
	perror(name) ;
	return("\n****Error Opening File****\n") ;
    }
    old_inbuf = inbuf; inbuf = (BUF *)0 ;
    old_input = infile ; infile = f ;
    old_name = thefile ; thefile = name ;
    old_lineno = lineno ; lineno = 1 ;
    old_colno = colno ; colno = 0 ;
    old_mode  = input_mode ; input_mode = 0 ;

    display_input((char *)0) ;

    input_mode = old_mode ;
    inbuf = old_inbuf ;
    infile = old_input ;
    thefile = old_name ;
    lineno = old_lineno ;
    colno = old_colno ;

    return("") ;
}

static char *exec_inject(char *name)
/*++++++++++++++++
.PURPOSE  Switch input to named buffer
.RETURNS  ""
.REMARKS  Call display_input from saved buffer
-----------------*/
{
  BUF *old_inbuf, *v ;

    /* Try to open file file... */
    v = var_find(name) ;
    if (!v) {
	fprintf(stderr, "\n****") ;
	perror(name) ;
	return("\n****Error .inject ****\n") ;
    }
    old_inbuf = inbuf; inbuf = v ;
    display_input((char *)0) ;
    inbuf = old_inbuf ;

    return("") ;
}

/*==================================================================
	Reading input and keeping BUFFERs of input data
 *==================================================================*/

static BUF *buf_alloc(int size)
/*++++++++++++++++
.PURPOSE  Allocate a new BUF, and returns
.RETURNS  New BUF / NULL
-----------------*/
{
  BUF *new; char *t ; int len;
    len = size ;
    if (len < 7) len = 7;
    len |= 7;
    t = malloc(len+1) ;
    if (! t) {
    	fprintf(_of, "\n\\verb\177**buf_alloc: Can't allocate %d bytes \177\n",
    	    size);
    	return((BUF *)0);
    }
    new = (BUF *)malloc(sizeof(BUF)) ;
    /* fprintf(_of, "....I allocated BUF %p\n", new); */
    new->size = len;
    new->used = 0;
    new->pos  = 0;
    new->text = t;
    return(new);
}

static int buf_clear(BUF *b)
/*++++++++++++++++
.PURPOSE  Make a buffer Empty
.RETURNS  0 / -1
-----------------*/
{
    if (!b) return(-1) ;
    (b)->used = (b)->pos = 0 ;
    if (b->text) b->text[0] = 0 ;
    return(0) ;
}


#define buf_kill(b)	buf_free(b), free(b)
#define buf_zero(b)	memset(b, 0, sizeof(BUF));

static int buf_free(BUF *b)
/*++++++++++++++++
.PURPOSE  Free a buffer
.RETURNS  0 / -1
-----------------*/
{
    if (b->text) free(b->text) ;
    b->text = (char *)0 ;
    b->size = b->used = b->pos = 0 ;
    return(0) ;
}

static int buf_expand(BUF *b, int bytes)
/*++++++++++++++++
.PURPOSE  Expand a buffer
.RETURNS  0 / -1
-----------------*/
{
  int i; char *p ;
    i = b->size + (bytes > 0 ? bytes : BUFINCR);
    if (!b->text) p = malloc(i) ;
    else p = realloc(b->text, i) ;
    if (!p) {
        fprintf(stderr, "**** (%s)", inloc()) ;
        perror("in buf_expand") ;
        return(-1) ;
    }
    b->size = i ;
    b->text = p ;
    return(0);
}

static int buf_checksize(BUF *b, int bytes)
/*++++++++++++++++
.PURPOSE  Expand buffer if necessary
.RETURNS  0 (OK)
-----------------*/
{
    if (b->size > bytes) return(0);
    return(buf_expand(b, 1+(bytes-b->size)));
}

static int buf_append(BUF *b, char *text, int len)
/*++++++++++++++++
.PURPOSE  Append a text in a buffer -- expand if necessary
.RETURNS  0 / -1
-----------------*/
{
  int i, more;
    if (len < 0) len = strlen(text);
    i = b->used + len ;
    if (i >= b->size) { 			/* Must realloc ...	*/
	more = i - b->size + 1 ;
	if (more < BUFINCR) more = BUFINCR ;
	if (buf_expand(b, more)) return(-1) ;
    }
    memcpy(b->text + b->used, text, len) ;
    b->text[i] = 0;
    b->used = i;
    return(0);
}

static int buf_achar(BUF *b, int c)
/*++++++++++++++++
.PURPOSE  Append a char to a buffer -- expand if necessary
.RETURNS  New BUF / NULL
-----------------*/
{
    if ((b->used+1) >= b->size)  	/* Must realloc ...	*/
	if(buf_expand(b, BUFINCR)) return(-1) ;
    b->text[b->used] = c ;
    b->used += 1 ;
    b->text[b->used] = 0 ;
    return(0) ;
}

static int buf_symbol(BUF *b, char *symbol)
/*++++++++++++++++
.PURPOSE  Append to specified buffer the meaning of the symbol
.RETURNS  Last read byte
.REMARKS  symbol starts by ${ and symbol[len] = '}'
	If symbol is NULL, assume $ read, get from input
-----------------*/
{
  int old_mode; BUF *old_collector ; char *old_begline ;
    old_begline = begline ;
    old_collector = collector ;
    old_mode = output_mode ;
    output_mode &= ~3 ;		/* Means: gather output into collector */
    output_mode |= 2 ;		/* Means: gather output into collector */
    collector = b ;
    match_symbol(symbol) ;
    collector = old_collector ;
    begline = old_begline ;
    output_mode = old_mode ;
    return(0) ;
}

static int next_char()
/*++++++++++++++++
.PURPOSE  "Peek", i.e. tell what's the next char without getting it
.RETURNS  The character / EOF
-----------------*/
{
  int c ;
    if (inbuf) {
    	if (inbuf->pos >= inbuf->used) return(EOF) ;
    	return(inbuf->text[inbuf->pos]) ;
    }
    c = getc(infile) ;
    ungetc(c, infile) ;
    /* if (c == '\n') c = ' ' ; */
    return(c) ;
}

static int get_char()
/*++++++++++++++++
.PURPOSE  Get the next char, either from stored or from stdin
.RETURNS  The character / EOF
-----------------*/
{
  int c;
    if (display_stop) {
	if (display_depth > display_till) return(display_stop) ;
	display_stop = 0 ;
    }
    if (inbuf) {
    	if (inbuf->pos >= inbuf->used) return(EOF) ;
    	c = inbuf->text[inbuf->pos] ;
    	inbuf->pos += 1 ;
    	return(c) ;
    }
    if (!infile) return(EOF) ;
    if (input_mode & MODE_nofile) return(EOF) ;
    c = getc(infile) ;
    if (c == '\n') { lineno++; colno = 0 ; }
    else colno++;
    /* if (c == '\n') { c = ' ' ; } */
    if (verbop) {	/* Save the input */
	if (!parsed_buf) parsed_buf = buf_alloc(BUFINCR) ;
	buf_achar(parsed_buf, c) ;
    }
    return(c);
}

static char *get_string()
/*++++++++++++++++
.PURPOSE  Get the next string as a new string
.RETURNS  Last read byte
.REMARKS  Reads until </...>
-----------------*/
{
  BUF but ;
    if (inbuf && inbuf->pos == 0) {	/* Short Cut */
    	inbuf->pos = inbuf->used ;
    	return(inbuf->text) ;
    }
    buf_zero(&but) ;
    buf_symbol(&but, "${...}") ;	/* Get Input */
    return(but.text) ;
}

static int get_until(BUF *b, int stop_char)
/*++++++++++++++++
.PURPOSE  Append to buf until a special character is found (e.g. quote)
.RETURNS  The last char read.
.REMARKS  The ending 'c' is appended to buffer.
-----------------*/
{
  int c ;
    while ((c = get_char()) != EOF) {
    	buf_achar(b, c) ;
    	if (c == stop_char) break ;
    }
    return(c) ;
}

static int get_brace(BUF *b)
/*++++++++++++++++
.PURPOSE  Append to buf until the mnatching brace
.RETURNS  The last char read.
.REMARKS  The ending brace is appended to buffer.
-----------------*/
{
  int depth, c ;
    depth = 1 ;
    while (depth) {
    	c = get_char() ;
    	if (c == EOF) break ;
    	buf_achar(b, c) ;
      	if (c == '{') { depth++ ; continue ; }
      	if (c == '}') { depth-- ; continue ; }
      	if (isquote(c)) get_until(b, c) ;
    }
    return(c) ;
}

static int get_brack(BUF *b)
/*++++++++++++++++
.PURPOSE  Append to buf until the matching bracket
.RETURNS  The last char read.
.REMARKS  The ending bracket is appended to buffer.
-----------------*/
{
  int depth, c ;
    depth = 1 ;
    while (depth) {
    	c = get_char() ;
    	if (c == EOF) break ;
    	buf_achar(b, c) ;
      	if (c == '[') { depth++ ; continue ; }
      	if (c == ']') { depth-- ; continue ; }
      	if (isquote(c)) get_until(b, c) ;
    }
    return(c) ;
}

static MARKER *mark_find(char *name)
/*++++++++++++++++
.PURPOSE  Find a named Marker
.RETURNS  The value (0 for non-existing marker) / the top marker (no name)
-----------------*/
{
  MARKER *v;
    v = top_mark ;
    while (v) {
	if (strcmp(v->name, name) == 0)
	   return(v) ;
	v = v->prev;
    }
    return((MARKER *)0) ;
}

static int mark_set(char *name, int value)
/*++++++++++++++++
.PURPOSE  Set a value in a named marker. Created if non-existant.
.RETURNS  The previous value (if existing)
-----------------*/
{
  int pval ; MARKER *v;
    if (!name) name = "" ;
    if (*name) v = mark_find(name) ;
    else v = (MARKER *)0 ;
    if (v) pval = v->value ;
    else {
	v = (MARKER *)malloc(sizeof(MARKER) + strlen(name) + 1) ;
	v->name = (char *)(v+1) ;
	strcpy(v->name, name) ;
	v->top_dic = top_dic ;
	v->prev = top_mark ; top_mark = v ;
	pval = 0 ;
    }
    v->value = value ;
    return(pval) ;
}

static int mark_kill(char *name)
/*++++++++++++++++
.PURPOSE  Remove a marker
.RETURNS  0 / -1
-----------------*/
{
  MARKER *p, *v;
    p = (MARKER *)0 ;
    for (v=top_mark; v && strcmp(name, v->name); p=v, v=v->prev) ;
    if (!v) return(-1) ;
    if (p) p->prev = v->prev ;
    else top_mark = v->prev ;
    v->prev = (MARKER *)0 ;
    free(v) ;
    return(0) ;
}


static BUF *var_find(char *name)
/*++++++++++++++++
.PURPOSE  Find a named buffer
.RETURNS  Buffer / NULL
-----------------*/
{
  NAMED_BUF *v;
  int i=0;
    for (v=top_var; v; v=v->prev) {
	if (strcmp(v->name, name) == 0)
	   return(&(v->buf)) ;
	i++;
    }
    if (i != count_var) fprintf(stderr,
      "****var_find(%s): only %d/%d named buffers found\n",
      name, i, count_var);
    return((BUF *)0) ;
}

static BUF *var_1find(char *name)
/*++++++++++++++++
.PURPOSE  Find a named buffer (from 1st word)
.RETURNS  Buffer / NULL
-----------------*/
{
  BUF *v;
  int len, x ;
    while (isspace(*name)) name++ ;
    for (len=0; isGraph(name[len]); len++) ;
    x = name[len]; name[len] = 0 ;
    v = var_find(name) ;
    name[len] = x ;
    return(v);
}

static BUF *var_set(char *name, char *value)
/*++++++++++++++++
.PURPOSE  Set a value in a named buffer. Created if non-existant.
.RETURNS  Buffer corresponding to name
-----------------*/
{
  static int nameno = 0 ;
  BUF *b ; NAMED_BUF *v;
  char buf[20] ;
    if (!name) sprintf(name=buf, "+#%d", ++nameno) ;
    b = var_find(name) ;
    if (!b) {		/* Create the named buffer */
	v = (NAMED_BUF *)malloc(sizeof(NAMED_BUF) + strlen(name) + 1) ;
	v->name = (char *)(v+1) ;
	strcpy(v->name, name) ;
	b = &(v->buf) ;
	buf_zero(b) ;
	v->prev = top_var ; top_var = v ;
	count_var++;
    }
    buf_clear(b) ;
    buf_append(b, value, -1) ;
    return(b) ;
}

static BUF *var_append(char *name, char *value)
/*++++++++++++++++
.PURPOSE  Append a string to a named buffer. Created if non-existant.
.RETURNS  Buffer corresponding to name
-----------------*/
{
  BUF *b ;
    if ((b = var_find(name))) buf_append(b, value, -1) ;
    else b = var_set(name, value) ;
    return(b) ;
}

static int var_kill(char *name)
/*++++++++++++++++
.PURPOSE  Remove a variable.
.RETURNS  0 / -1
-----------------*/
{
  NAMED_BUF *p, *v;
    p = (NAMED_BUF *)0 ;
    for (v=top_var; v && strcmp(name, v->name); p=v, v=v->prev) ;
    if (!v) return(-1) ;
    buf_free(&(v->buf)) ;
    v->name[0] = 0 ;
    if (p) p->prev = v->prev ;
    else top_var = v->prev ;
    v->prev = (NAMED_BUF *)0 ;
    free(v) ;
    count_var -= 1;
    return(0) ;
}

static BUF *save_tag(char *text)
/*++++++++++++++++
.PURPOSE  Save a tag from its ID
.RETURNS  The buffer where it's saved
.REMARKS  The name is =ID ('=' char followed by the contents of ID)
-----------------*/
{
  char name[256] ;
  char *p ;
  int k ;

    k = strarg(text, "ID") ;
    if (!text[k]) return((BUF *)0) ;
    p = text + k + 2 ;
    if (*p != '=') return((BUF *)0) ;
    p++ ;		/* Skip the = sign */
    /* Get the name as string of length k */
    if (isquote(*p)) { k = strquote(p); p++; k-- ; }
    if (k > (sizeof(name)-2)) k = sizeof(name)-2 ;
    strncpy(name+1, p, k) ;
    name[k+1] = 0 ;
    name[0] = '=' ;

    return(var_set(name, text)) ;
}

/*==================================================================
		Output Functions
 *==================================================================*/

int putchar(int c)
/*++++++++++++++++
.PURPOSE  Output a character to output file
.RETURNS  Character issued / EOF / 0 (not issued, reformatting)
.REMARKS  An 8-bit character is issued according to otex.
	Strange feature: when c = -19 (iacute), the result of isspace
	is sometimes true on Solaris ???
-----------------*/
{
  int status = 0 ;
  static char pending_space ;

    c &= 0xff ;		/* !! EXPERIENCED PROBLEMS ON SOLARIS !!!  */
    if ((output_mode&MODE_nl2cr) && (c == '\n'))
	c = '\r';
    if (c == '\r') { 	/* Be at left margin 			   */
	pending_space = 0 ;
	if (colpos) putc('\n', _of) ;
	colpos = 0 ;
	return(0) ;
    }
    if (output_mode&(MODE_force|MODE_verb)) ;
    else if ((c == '\n') && (!(output_mode&MODE_obey))) {/* don't force \n  */
	if (colpos > 0) pending_space = ' ';
	return(0) ;
    }
    else if (isspace(c) && owidth) { 		/* Reformatting... */
	/* fprintf(_of, "(space for %o)", c) ;  */
        pending_space = colpos >= owidth ? '\n' : ' ' ;
        return(0) ;
    }
    if (pending_space) {
        putc(pending_space, _of) ;
        if (pending_space == '\n') {
	    colpos = 0 ;
    	    for (colpos=0; colpos<indent; colpos++) putc(' ', _of) ;
	}
	else colpos += 1 ;
        pending_space = 0 ;
    }
    if (c == '\n') colpos = 0 ;
    else colpos += 1 ;
    if (c&128)  {	/* 8-bit extensions (Latin assumed) */
        if (char8b) fprintf(_of, "%s", char8b[c&127]);
    	else if (latin2dos) fprintf(_of, "%s", latin2dos[c&127]) ;
    	else putc(c, _of) ;
    }
    else if (output_mode&(MODE_upper|MODE_lower) && isalpha(c)) {
    	if (output_mode&MODE_upper) c = toupper(c) ;
    	else c = tolower(c) ;
    	putc(c, _of) ;
    }
    else putc(c, _of) ;

    /* Beginning of lines in OBEY mode: insert e.g. # */
    if ((output_mode&MODE_obey) && begline && (colpos == 0))
        fprintf(_of, "%s", begline), colpos = strlen(begline) ;

    return(status) ;
}

static int putstring(char *s)
/*++++++++++++++++
.PURPOSE  Output a String to _of
.RETURNS  cf printf
-----------------*/
{
  int status = 0 ;
  char *p ;
    if (owidth || (output_mode&(~1))) {
	for (p=s; *p; p++) status += putchar(*p) ;
	return(status) ;
    }
    for (p=s; *p; p++) {
        if (*p == '\n') colpos = 0 ;
        else colpos += 1 ;
        if ((*p)&128)  		/* 8-bit extensions (Latin assumed) */
    	     putchar(*p) ;	/* Take care of conversion */
    	else fputc(*p, _of) ;
    }
    return(status) ;
}

static int putbuffer(char *s, int len)
/*++++++++++++++++
.PURPOSE  Output a text
.RETURNS  Number of bytes issued.
-----------------*/
{
  int status = 0 ;
  char *p, *e ;
    if (owidth || (output_mode&(~1))) {
	for (p=s, e=p+len; p<e; p++) status += putchar(*p) ;
	return(status) ;
    }
    status = fwrite(s, 1, len, _of) ;
    for (p=s, e=p+len; p<e; p++) {
        if (*p == '\n') colpos = 0 ;
        else colpos += 1 ;
    }
    return(status) ;
}

static int PutChar(int c)
/*++++++++++++++++
.PURPOSE  Output a character
.RETURNS  Character issued / EOF / 0 (not issued, reformatting)
.REMARKS  An 8-bit character is issued according to otex.
-----------------*/
{
  int status = 0, i, x ;
    if ((output_mode&MODE_nl2cr) && (c == '\n'))
	c = '\r';
    if (output_mode&1) status = putchar(c) ;
    if (output_mode&2) {
        if (c == '\r') { 	/* Be at left margin	*/
	    for (i=collector->used-1; i>=0; i--) {
		x = collector->text[i] ;
		if (x == '\n') { c=0; break; }
		if (!isspace(x)) break ;
	    }
	    ++i ;
	    if (collector->text) collector->text[i] = c ;
	    collector->used = i ;
	    if (!c) return(status) ;
	    c = '\n' ;
        }
	x = ' ';		/* Last byte */
        if (!collector) collector = buf_alloc(BUFINCR) ;
	else {
	    i=collector->used-1;
	    if (i>=0) x = collector->text[i];
	}
        if (output_mode&(MODE_force|MODE_verb|MODE_obey)) ;
#if 0	/* NO !!! We have to respect the definitions!! */
	else if (isspace(c) && isspace(x)) 	/* Only a single space */
	    return(0);
#endif
        else if (output_mode&(MODE_upper|MODE_lower))
            c = output_mode&MODE_upper ? toupper(c) : tolower(c) ;
    	buf_achar(collector, c) ;
    }
    return(status) ;
}

static int PutString(char *s)
/*++++++++++++++++
.PURPOSE  Output a String
.RETURNS  cf printf
-----------------*/
{
  int len ;
  int status = 0 ;
  char *p ;
    if (output_mode&1) status = putstring(s) ;
    if (output_mode&2) {
        len = strlen(s) ;
        if (!collector) collector = buf_alloc(len) ;
        if (output_mode&(MODE_upper|MODE_lower))
             p = trstr(s, output_mode&MODE_upper) ;
        else p = s ;
    	buf_append(collector, p, len) ;
    }
    return(status) ;
}

static int PutText(char *s, int len)
/*++++++++++++++++
.PURPOSE  Output a text
.RETURNS  Number of bytes issued.
-----------------*/
{
  int status = 0 ;
  char *p ;
    if (output_mode&1) status = putbuffer(s, len) ;
    if (output_mode&2)  {
        if (!collector) collector = buf_alloc(len) ;
        if (output_mode&(MODE_upper|MODE_lower))
             p = trstr(s, output_mode&MODE_upper) ;
        else p = s ;
    	buf_append(collector, p, len) ;
    }
    return(status) ;
}

static int PutTextUnquoted(char *s, int slen)
/*++++++++++++++++
.PURPOSE  Output a text, removing the basic &amp; &lt; &gt; &apos; &quot;
.RETURNS  Number of bytes issued.
-----------------*/
{
  static char conv[] = "amp;&lt;<gt;>apos;'quot;\"minus;-hyphen;-ndash;-" ;
  char *p, *e, *c, *t=0;
  int i, status ;
    status = 0 ;
    p = s ; e = p + slen ;
    while (p < e) {
	if (*p == '&') {
	    p++ ;
	    for (c=conv; *c; c++) {
		/* i set to non-zero when 2 chars differ */
		for (t=p, i=0; *c != ';'; c++, t++) i |= (*c ^ *t) ;
		if ((i == 0) && (*t == *c)) break ;
		c++ ;
	    }
	    if ((*c == ';') && (t < e)) { PutText(c+1, 1) ; p = t+1 ; }
	    else                        { PutText("&", 1) ; p++ ; }
	    status++ ;
	    continue ;
	}
	for (t=p; (t<e) && (*t != '&'); t++) ;
	PutText(p, i=t-p) ;
	p = t ;
	status += i ;
    }
    return(status) ;
}

static char *tr_entity(char *text)
/*++++++++++++++++
.PURPOSE  Convert an entity
.RETURNS  Equivalent
.REMARKS  Added V3.5
-----------------*/
{
  static char eqs[32];
  char buf[30], *p, *t;
  int i;
    i = 0 ;
    buf[i++] = '&' ;
    t = text; if (*t == '&') t++;
    while (isTagChar(*t) && (i < sizeof(buf)-3)) {
	buf[i++] = *(t++);
    }
    buf[i++] = ';' ;
    buf[i] = 0 ;

    /* Can it be a simple number (Latin) ? */
    if (buf[1] == '#') {
	i = atoi(buf+2) ;
	if ((i&(~127)) == 128) {	/* Range  128..255 */
	    eqs[0] = i;
	    eqs[1] = 0;
	    return(eqs) ;
	}
    }

    /* Get the Translation */
    if ((p = lookup(top_dic, buf))) {
    	p += strloc(p, '=');
    	if (*p) p++ ; 			/* Skip the = sign */
        return(p);
    }

    /* Don't know what this is -- take defaults according to tex/tex2 */
    i = strloc(buf, ';') ; buf[i+1] = 0 ;
    if (otex&3) buf[0] = '{', buf[i] = '}' ;
    else buf[0] = '[', buf[i] = ']' ;
    if ((otex&3) == 1) { eqs[0] = buf[0]; eqs[1] = '\\'; strcpy(eqs+2, buf+1); }
    else strcpy(eqs, buf);
    return(eqs) ;
}

int sgml_unquote(char *text)
/*++++++++++++++++
.PURPOSE  Convert the text (inplace remove the &...;)
.RETURNS  Length of text.
-----------------*/
{
  char buf[30], *p, *s, *t ;
  int i, c=0;
    for (s=t= text; *t; t++) {
        if (*t != '&') { *(s++) = *t ; continue; }
	/* Here we start by & -- Look for entity */
        i = 0 ;
        buf[i++] = *(t++) ; 	/* Insert the '&' */
	while(isGraph(*t)) {
	    c = *(t++) ;
    	    buf[i++] = c ;
    	    if (c == ';') break ;
    	    if (c == '&') { --i; --t; break; }	/* e.g. &amp&amp; */
    	    if (i >= (sizeof(buf)-2)) break ;
    	}
	if (c != ';') buf[i++] = ';' ;
	--t ;					/* for loop t++   */
        buf[i] = 0 ;

        /* Can it be a simple number (Latin) ? */
        if (buf[1] == '#') { int n;
	    n = atoi(buf+2);
	    if (n<256) { *(s++) = n; continue ; }
	}

        /* Get the Translation -- single letter only */
        if ((p = lookup(top_dic, buf))) {
    	    p += strloc(p, '=');
    	    if (*p && p[1] && (!p[2])) {
	        *(s++) = p[1];
		continue ;
	    }
	}
	strcpy(s, buf); s += strlen(buf) ;
    }
    *s = 0 ;

    return(s-text) ;
}

static int PutStringUnquoted(char *s)
/*++++++++++++++++
.PURPOSE  Output a text, removing the basic &amp; &lt; &gt; &apos; &quot;
.RETURNS  Number of bytes issued.
-----------------*/
{
    return(PutTextUnquoted(s, strlen(s))) ;
}

/*==================================================================
		Manipulation of Dictionaries
 *==================================================================*/

static int t1cmp(char **e1, char **e2)
/*++++++++++++++++
.PURPOSE  SORT function to sort rows of a Dictionary
.RETURNS  -1 / 0 / 1
-----------------*/
{
#define Return(v) return(v)
/* #define Return(v) { printf("=>%d\n", v) ; return(v) ; } */
  char *p1, *p2 ; int diff ;
  p1 = *e1 ; p2 = *e2 ;
    if (*p1 != *p2) return(*p1 - *p2) ;
    switch(*p1) {
      case '&':	/* Entities &bablabla; */
        while ((*p1 == *p2) && (*p1 != ';') && (*p1)) p1++, p2++ ;
        if (*p1 == *p2) Return(0) ;
        if (*p1 == ';') Return(- *p2) ;
        if (*p2 == ';') Return(*p1) ;
        Return(*p1 - *p2) ;
      case '<':	/* Tags, case insensitive (Mod. V3.5) */
        p1++, p2++ ;  diff = 0 ;
        while ((diff == 0) && isTagChar(*p1) && isTagChar(*p2)) {
            diff = toupper(*p1) - toupper(*p2) ;
            p1++; p2++ ;
        }
        if (diff) Return(diff) ;
        if (*p1 == *p2) Return(0) ;
        if (*p1 == '>') Return(isTagChar(*p2) ? - *p2 : 0);
        if (*p2 == '>') Return(isTagChar(*p1) ?   *p1 : 0);
        Return(*p1 - *p2) ;
      case '$':	/* Compare ${.action}, stop at space or closing */
        p1++, p2++ ;  diff = 0 ;
        while ((diff == 0) && isGraph(*p1) && isGraph(*p2) /* {{ */
           && (*p1 != '}') && (*p2 != '}')) {
            diff = *p1 - *p2 ;
            p1++; p2++ ;
        }
        if (diff) Return(diff) ;
        if (*p1 == *p2) Return(0) ;
        /*{*/ if (*p1 == '}') Return(isspace(*p2) ? 0 : - *p2 );
        /*{*/ if (*p2 == '}') Return(isspace(*p1) ? 0 :   *p1 );
        Return(*p1 - *p2) ;
    }
    while ((*p1 == *p2) && (*p1)) p1++, p2++ ;
    Return(*p1 - *p2) ;
}

static DIC *t1new(int items)
/*++++++++++++++++
.PURPOSE  Allocate a new dictionnary
.RETURNS  The translation
-----------------*/
{
  DIC *new;
    new = (DIC *)malloc(sizeof(DIC) + (items-2)*sizeof(char *));
    new->buf = (char *)0;
    new->prev = 0;
    new->items = items;
    new->flags = 0;
    return(new);
}

static void t1kill(DIC *dic)
/*++++++++++++++++
.PURPOSE  Delete a Dic
.RETURNS  ---
-----------------*/
{
    dic->buf = (char *)0;
    dic->items = 0;
    free(dic);
    return;
}

static char *lookup(DIC *dic, char *s)
/*++++++++++++++++
.PURPOSE  Find out correspondance in e.g. texdef_dic dic
.RETURNS  The full matching line of the Dic
.REMARKS  Returns the first definition in case of multiple definitions
	Keep the result in static lookup_entry.
-----------------*/
{
  int i, stop_here; char **a, *arg;
    a = lookup_entry = (char **) 0;
    arg = s ;
    stop_here = (*s == '<' ? DIC_IGNORE : 0) ;
    while (dic) {
    	if (dic->flags&DIC_SORTED) {	/* SORTED Dic.	*/
	    a = bsearch(&arg, dic->entries, dic->items,
	    	sizeof(char *), (SORT_FCT)t1cmp) ;
	    /* Multiple definitions: go to first one */
	    if (a && (dic->flags&DIC_MDEF)) {
		while ((--a >= dic->entries) && (t1cmp(a, &arg) == 0)) ;
		++a;
	    }
	}
    	else {
    	    for ( i=dic->items, a=dic->entries; --i >= 0; a++) {
    	        if (t1cmp(a, &arg) == 0) break ;
	    }
	    if (i < 0) a = (char **)0 ;
	}
	if ((dic->flags&stop_here) && (!a))
	    return("<...>=${.ignore}${...}") ;	/* */
	if (a) break;
	dic = dic->prev ;
    }

    /* Take Care of Possible Synonym, like <tag>:<synonym> */
    if (a && (*s == '<')) {
    	arg = *a + strloc(*a, '>') ;
    	if (arg[0] && (arg[1] == ':') && (arg[2] == '<')) {
    	    arg += 2 ;		/* New tag to locate */
    	    arg = lookup(dic, arg) ;
    	    a = &arg ;
    	}
    }
    /* Save the result in statics (V3.0) */
    if (a) {
	lookup_entry = a;
	lookup_last  = dic->entries + dic->items;
	return(*a);
    }
    return((char *)0) ;
}

static char *lookup_with_attr(DIC *dic, char *s)
/*++++++++++++++++
.PURPOSE  Find out correspondance with attribute
.RETURNS  The full matching line of the Dic
.REMARKS  Returns the first definition in case of multiple definitions
	Keep the result in static lookup_entry.
-----------------*/
{
  char *p, *arg, *px, x;
  int found;
    p = lookup(dic, s);
    if (!p) return(p);
    if (*p != '<') return(p);
    if (!lookup_entry) return(p);

    /* The dic entry contains an attribute.
       Verifies this attribute exists in input string
       Note that p = *lookup_entry  -- unless it's a default definition!
       Also, definitions without argument are always the last one, e.g.
       <ADDRESS type="publisher">=...
       <ADDRESS>=${...}
    */

  try_Next:
    for (p = *lookup_entry; isGraph(*p) && (*p != '>'); p++) ;
    if (!isspace(*p)) return(*lookup_entry);
    found = 1;

    while (isspace(*p) && found) {	/* Look for next attribute */
	p++;
	if (isspace(*p)) continue;
        /* Locate the end of the attribute */
        for (px=p; *px; px++) {
	    if (isquote(*px)) { px += strquote(px); continue; }
	    if (isspace(*px) || (*px == '>')) break;
        }
        x = *px; *px = 0;
	if (!strstr(s, p)) found = 0;
        *px = x;
    }
    if (found) return(*lookup_entry);
    if (++lookup_entry < lookup_last) {
        arg = s;
	if (t1cmp(lookup_entry, &arg)) return((char *)0);
	goto try_Next;
    }
    return((char *)0) ;
}

/*==================================================================
		Loading Dictionaries
 *==================================================================*/

static void load_dic(char *list, int flags)
/*++++++++++++++++
.PURPOSE  Load the Dictionary from the buffer
.RETURNS  ---
.REMARKS  Only lines starting by < or &
-----------------*/
{
static char badly_sorted[] = "\
\\verb\177****Basic Definitions badly sorted '%s' vs '%s'\177\n\
";
  int i, i2, len, counts, mdef;
  char *p, *m, *e, **a, x;
  DIC *t1 ;

#if DEBUG
    printf("....Entered load_dic:\n%s\n----(End of Definitions)\n", list) ;
#endif
    len = strlen(list);

    /* Count the number of valid definitions */
#if 0
    counts = 0 ;
    for (p=list, e=p+len; p<e; p += strlen(p)+1) {
	i = strloc(p, '\n'); p[i] = 0;
	if (!strchr(c1dic,*p)) {	/* Can be a continuation...	*/
	    if (p > list) p[-1] = '\n';
	    continue;
	}
	if (*p == '#') continue;

	/* if ((*p == '<') || (*p == '&') || (*p == '$')) */
	counts += 1 ;
    }
#else
    counts = 1;
    p=list, e=p+len;
    while (p<e) {
	if (*p == '{') {
	    ++p;
	    i = strbrace(p) ;
	    if (p[i] != '}') {
		fflush(stdout) ;
		i2 = i < 40 ? i : 40;
		x = p[i2]; p[i2]=0;
		fprintf(stderr, "\n****Nonclosed brace ? %s\n",p-1) ;
		p[i2]=x;
	    }
	    p += i ;
	    continue ;
	}
	if (*p == '\n') {
	    if (strchr(c1dic,p[1]))	{	/* An element for the dic */
	        counts++ ;
	        *p = 0 ;
	    }
	}
	p++ ;
    }
#endif

    /* Allocate the dictionnary, and set up the pointers */
    t1 = t1new(counts);
    t1->buf  = list;
    t1->prev = top_dic ;
    a = t1->entries ;

    /* Enter the lines of definition in the dictionary */
    for (p=list, e=p+len; p<e; p += strlen(p)+1) {
	if (!strchr(c1dic,*p))	continue ;
    	if (*p == '#') 		continue ;
	if (!*p) continue ;
	/* if ((*p == '<') || (*p == '&')) */
	*(a++) = p ;
	/* Lines ending with \ : ignore until non-blank */
#if 0	/* Merge \\n into a single line */
	p += strloc(p, '\\') ; m = p ;
	while ((p<e) && (*p == '\\')) {
	    if (p[1] == '\n') for (++p; isspace(*p);  p++) ;
	    else *(m++) = *(p++) ;
	    while(*p && (*p != '\\')) *(m++) = *(p++) ;
	}
	while (*p) *(m++) = *(p++) ;
	*(m--) = 0 ;			/* m -> last character  */
	while ((p<e) && (*m == '\\')) {	/* Merge lines ending with \ */
	    ++p ;			/* Skip the EOL */
	    while (isspace(*p)) p++ ;	/* Skip starting blanks */
	    while (*p) *(m++) = *(p++) ;
	    *(m--) = 0 ; 		/* m -> last character  */
	}
#else	/* Merge \n\ into newline, \ in skip blanks */
	p += strloc(p, '\\') ;
	while (*p) {
	    if ((p[1] == 'n') && (p[2] == '\\') && ((!p[3])||(p[3]=='\n')))
		break ;
	    if ((!p[1]) || (p[1] == '\n')) break ;
	    p++;
	    p += strloc(p, '\\') ;
	}
	if (!*p) continue ;
	m = p ;				/* Target string (transformed \) */
	/* There are lines to merge -- do it ! */
	while ((p<e) && (*p == '\\')) {
	    if ((p[1] == 'n') && (p[2] == '\\') && ((!p[3])||(p[3]=='\n')))
		*(m++) = '\n', p += 4 ;
	    else if ((!p[1]) || (p[1] == '\n'))
		for (p+=2; isspace(*p); p++) ;
	    else *(m++) = *(p++) ;
	    while (*p && (*p != '\\')) *(m++) = *(p++) ;
	}
	*m = 0 ;
#endif
    }
    t1->items = (a - t1->entries) ;

    /* Sort the Dictionnary if there are many (> 20) definitions */
    if (t1->items > 20) {
        mdef = 0;
    	if (flags&DIC_SORTED) {	/* Just verify it's sorted */
    	    /* printf("....Compare %d items\n", t1->items) ;  */
	    for (i2 = t1->items, a = t1->entries; --i2 > 0; a++) {
		i = t1cmp(a, a+1);
		if (i<0) continue;
		if ((i==0)&&(flags&DIC_MDEF)) { mdef = DIC_MDEF; continue; }
		fprintf(_of, badly_sorted, a[0], a[1]);
    	    }
        }
    	else {
	    qsort(t1->entries, t1->items, sizeof(char *), (SORT_FCT)t1cmp) ;
	    /* Verify the DIC_MDEF */
	    if (flags&DIC_MDEF)
	      for (i2 = t1->items, a = t1->entries; --i2 > 0; a++) {
		if (t1cmp(a, a+1) == 0) { mdef = DIC_MDEF; break; }
	    }
	}
	flags &= ~DIC_MDEF;
    	flags |= mdef ;
    	flags |= DIC_SORTED ;
	/* Verify -- are there multiple definitions ? */
    }
    t1->flags = flags ;
#if DEBUG
    printf("....Loaded dic with %d items\n", t1->items) ;
#endif

    top_dic = t1 ;
    return ;
}

static char *print_var()
/*++++++++++++++++
.PURPOSE  Display the list of Variables
.RETURNS  NULL
-----------------*/
{
  NAMED_BUF *v ;
    if (otex) fprintf(_of, "\\begin{verbatim}\n") ;
    fprintf(_of, "========List of Known Variables:\n") ;
    for (v=top_var; v; v=v->prev) fprintf(_of,
	"----Contents of %16s: \"%s\"\n", v->name,
	v->buf.text ? v->buf.text : "") ;
    return("") ;
}


static char *print_tags(why)
/*++++++++++++++++
.PURPOSE  Display the known tags
.RETURNS  NULL
-----------------*/
  char *why ; 	/* IN: Title */
{
  int n, i ; char **a; DIC *t1;

    if (output_mode == 0) return("") ;
    if (otex) fprintf(_of, "\\begin{verbatim}\n") ;
    fprintf(_of, "========List of Tags from %s options\n", why);
#if 0
    for (a=aaenv, i=strlen(tag2html); *a; a++) {
	if (strncmp(tag2html, *a, i) == 0) fprintf(_of, "$%s\t\n", *a);
    }
#endif
    for (n=0, t1 = top_dic ; t1 ; t1 = t1->prev, n--) {
	fprintf(_of, "....(Dictionnary #%d)....\n", n);
    	for (i=t1->items, a=t1->entries; --i>=0; a++)
    	    fprintf(_of, "%s\n", *a);
    }

    fprintf(_of, "========(End of list of Tags)===========-------------\n");
    if (otex) fprintf(_of, "\\end{verbatim}\n") ;

    /* Write out the list of Variables */
    print_var() ;

    return((char *)0);
}

/*==================================================================*
		Special Actions
 *==================================================================*/

static char *comment_text()
/*++++++++++++++++
.PURPOSE  Get the current comment
.RETURNS  The comment or ""
-----------------*/
{
    if (!comment_buf) return("") ;
    if (!comment_buf->text) return("") ;
    return(comment_buf->text) ;
}


static char *tagname(char *text)
/*++++++++++++++++
.PURPOSE  Get the tag name (in lowercase)
.RETURNS  The name (saved in a static buffer)
.REMARKS  If texst is null, use current_tag
	Within a comment, tag_name = full comment.
-----------------*/
{
  static char buf[LEN_TAG+1] ;
  int i ; char *p ;
    if (verbop&2) {
	printf("....tagname(%p=", text); fflush(stdout);
	if (text) printf("'%s'", text); 
	else {
	    printf("=>%p=", current_tag); fflush(stdout);
	    printf("'%s'", current_tag);
	}
	printf(")\n");
	fflush(stdout);
    }
    p = text ;
    if (!p)  p = "" ;
    if (!*p) p = current_tag ;
    if (!p) return ("") ;
    if (p == comment_tag) return(comment_text()) ;
    if (*p == '<') p++ ;
    for (i=0; (i<sizeof(buf)-1) && isGraph(*p) && (*p != '>'); p++)
	buf[i++] = tolower(*p) ;
    buf[i] = 0 ;
    if (verbop&2) {
	printf("        ==> %p='%s'\n", buf, buf);
    }
    return(buf) ;
}

static char *fulltag(char *text)
/*++++++++++++++++
.PURPOSE  Get the full tag (< and > excluded)
.RETURNS  The full tag (saved in a static buffer)
.REMARKS  If text is null, use current_tag
-----------------*/
{
  static char *abuf ; static size_t mbuf ;
  int i ; char *e, *p ;
    p = text ;
    if (!p)  p = "" ;
    if (!*p) p = current_tag ;
    if (!p) return ("") ;
    if (p == comment_tag) return(comment_text()) ;
    if ((i=strlen(p)) >= mbuf) {
	mbuf = 1 + (i|255) ;
	if (abuf) free(abuf) ;
	abuf = malloc(mbuf) ;
    }

    /* Remove the < > */
    for (e=p+i-1; (e>p) && isspace(*e); e--) ;
    if (*p == '<') {
	p++ ;
	if (*e == '>') e-- ;
    }
    strncpy(abuf, p, i=1+(e-p)) ;
    abuf[i] = 0 ;
    return(abuf) ;
}

static char *def_tag(char *text)
/*++++++++++++++++
.PURPOSE  Triggered by ${.def}, load a new dictionary from text
.RETURNS  ""
-----------------*/
{
  int i, sorted, nl ;
  BUF abuf ;
  char *p, *opt_text ;	/* Where to look for options 	*/
    if (verbop) printf("----Entered def_tag(\"%s\")\n", text) ;
    buf_zero(&abuf) ;
    if ((text[0] == '$') && (text[1] == '{' /*}*/)) { int old_mode ;
	old_mode = output_mode ;
	output_mode = MODE_verb|MODE_buf ;
	buf_symbol(&abuf, text) ;
	output_mode = old_mode ;
    }
    else abuf.text = strdup(text) ;
    sorted = 0 ;
    /* Options 'ignore' 'sort' may exist in the text */
    opt_text = (char *)0 ;
    if (strncasecmp(current_tag, "<def ", 4) == 0)
	opt_text = current_tag;
    else if (isalpha(*text))
	opt_text = text ;
    if (opt_text) nl = strloc(opt_text, '\n') ;
    if (opt_text) {
	i = strarg(opt_text, "sorted") ;
	if (opt_text[i] && (i < nl)) {
	    sorted = DIC_SORTED;
	    p = opt_text+i+6 ;
	    if (p[0] == '=') sorted = atoi(p+1) ;
	}
	i = strarg(opt_text, "ignore") ;
	if (opt_text[i] && (i < nl)) {
	    sorted = DIC_IGNORE ;
	    p = opt_text+i+6 ;
	    if (p[0] == '=') sorted = atoi(p+1) ;
	}
    }
    load_dic(abuf.text, sorted|DIC_MDEF) ;
    return("") ;
}

int subreg(char *cexp, char *replace, char *input, int max_replace)
/*++++++++++++++++
.PURPOSE  Execute a Substitute of a Regular Expression
.RETURNS  Number of Substitutions
-----------------*/
{
  char *p, *a, *r ;
  int saved_circ, saved_mode, n, c;

    if (verbop&2) printf("----subreg(/%s/ \"%s\", max_replace=%d)\n",
            cexp, replace, max_replace) ;

    /* Get First the string to substitute into a buffer */
    a = (char *)0 ;
    if (!input) a = input = get_string() ;

    saved_circ = circf ; circf = 0 ;
    saved_mode = output_mode ; output_mode |= MODE_force ;
    for (p = input, n=0 ; *p && (circf == 0) && (n < max_replace); n++) {
        if (verbop&4)
            printf("....Before step: p=%p'%s', loc1=%p, loc2=%p, circf=%d\n",
            p,p,loc1,loc2,circf);
    	if (!step(p, cexp)) break ;
        if (verbop&4) {
            printf(".... After step: p=%p'%s', loc1=%p, loc2=%p, circf=%d\n",
            p,p,loc1,loc2,circf);
            oterm = getch() ;
            if (oterm == 3) exit(1) ;
        }
    	PutText(p, loc1-p) ;
	for (r=replace; *r && (*r != '/'); r++) {
	    if (*r != '\\') { PutChar(*r); continue; }
	    ++r ;
	    if (isdigit(*r)) {	/* Replace \1 \2 ... */
		c = *r - 1;
		if ((c >= 0) && (c < nbra)) {
		    PutText(braslist[c], braelist[c] - braslist[c]) ;
		    continue ;
		}
	    }
	    if (!*r) { PutChar('\\') ; continue ; }
	    switch(c = *r) {
	      case 'n': c = '\n'; break ;
	      case 't': c = '\t'; break ;
	      case 'v': c = '\v'; break ;
	      case 'r': c = '\r'; break ;
	      case 'f': c = '\f'; break ;
	    }
	    PutChar(c) ;
	}
	if (*r == '/') {	/* Ckeck for global change */
	    if (*++r == 'g') --n ;
	}
    	p = loc2 ;
    }
    if (*p) PutString(p) ;

    output_mode = saved_mode;
    circf = saved_circ ;
    if (a) free(a) ;
    return(n);
}

static int exec_subs(char *arg, int max_replace)
/*++++++++++++++++
.PURPOSE  Triggered by ${.sub /reg_exp/ "substitute_string"},
.RETURNS  Number of substitutions,
.REMARKS  Executes on buffer, or on variable
-----------------*/
{
  static BUF subs_buf ;		/* Buffer to collect the result */
  static char *anil;
  char **a, *cexp, *subs, *input, *old_begline=0;
  BUF *b, *old_collector=0;
  int n, old_mode=0;

    /* Find Out the 3 strings: */
    if (verbop&2) printf("....exec_subs('%s', %d)\n", arg, max_replace) ;

    /* 0. Initialisations */
    a = &anil;

    /* 1. The Regular Expression: */
    if (*arg != '\05')	{	/* non-Compiled Expression ? */
    	fprintf(_of,
    	    "****non-compiled expression in exec_subs(\"%s\", %d)\n",
    	    arg, max_replace) ;
        cexp = compiled_reg ;
        n = *(arg++) ;		/* Keep the left/ of regexp */
        arg += strloc(arg, n) ;
    }
    else cexp = compiled_reg ;

    /* Get the strings for replacement */
    ++arg ;
    while (isspace(*arg)) arg++ ;
    if (*arg == '/') {		/* Skip the / starting the replacement */
	subs = ++arg ;
	arg += len_expr(subs) ;
	while (isGraph(*arg)) arg++ ;
    }
    else subs = "" ;
    while (isspace(*arg)) arg++ ;
    if ((*arg == '$') || isquote(*arg)) {	/* Text on which to operate */
        a = getargs(arg, 1, 1) ;
        input = a[0] ;
	b = (BUF *)0 ;
    }
    else {
	b = var_1find(arg) ;
	if (!b) {
	    fprintf(stderr, "++++Can't find variable '%s' for substitution\n",
	    arg) ;
	    return(0) ;
	}
	input = b->text ;
        old_begline = begline ;
        old_collector = collector ;
        old_mode = output_mode ;
        output_mode &= ~3 ;	/* Means: gather output into collector */
        output_mode |= 2 ;	/* Means: gather output into collector */
	collector = &subs_buf ;
	buf_clear(&subs_buf) ;
    }
    if (!input) input = "";

    n = subreg(cexp, subs, input, max_replace) ;

    if (b) {			/* Copy the result of substitution */
	buf_clear(b) ;
	buf_append(b, subs_buf.text, -1) ;
        collector = old_collector ;
        begline = old_begline ;
        output_mode = old_mode ;
    }

    /* Release allocated space */
    if (a[0]) free(a[0]);

    return(n) ;
}

static char *exec_sub(char *arg)
{
    exec_subs(arg, 1) ;
    return("") ;
}
static char *exec_gsub(char *arg)
{
    exec_subs(arg, 9999) ;
    return("") ;
}

static int exec_fmt(char *fmt)
/*++++++++++++++++
.PURPOSE  Triggered by ${%-format}, reformat the input
.RETURNS  0
.REMARKS  Executes on buffer, or on variable
-----------------*/
{
  static BUF var_buf ;		/* Buffer to collect the result */
  char *old_begline, *p;
  BUF *old_collector;
  int n, old_mode;
  double value;

    /* 1. Get the data */
    old_begline = begline ;
    old_collector = collector ;
    old_mode = output_mode ;
    output_mode &= ~3 ;	/* Means: gather output into collector */
    output_mode |= 2 ;	/* Means: gather output into collector */
    collector = &var_buf ;
    buf_clear(collector) ;
    display_input((char *)0);
    collector = old_collector ;
    begline = old_begline ;
    output_mode = old_mode ;

    /* Now interprets the Data */
    buf_checksize(&var_buf, 256);
    p = var_buf.text;		/* Result to print */
    switch(tolower(fmt[strlen(fmt)-1])) {
      case 'f': case 'e': case 'g':
	value = a2f(var_buf.text);
	sprintf(var_buf.text, fmt, value);
	break;
      case 's':		/* String */
	n = atoi(fmt+1);
	if (n<0) n = -n;
	if (n) {
	    buf_checksize(&var_buf, n+2+var_buf.used);
	    sprintf(p=var_buf.text+var_buf.used+1, fmt, var_buf.text);
	}
	else p = var_buf.text;
	break;
      default: 		/* Integer */
	value = a2f(var_buf.text);
	n = value;
	if ((value-n) != 0) {
	    fprintf(stderr, "++++Rounded value: %g\n", value);
	    value = rint(value);
	    n = value;
	}
	sprintf(var_buf.text, fmt, n);
    }
    PutString(p);
    return(0);
}


#if 0
static char *collect_until(char *tag)
/*++++++++++++++++
.PURPOSE  Triggered by ${.collect </end>}, collects data in input
.RETURNS  The text until tag is found
.REMARKS  tag is read. A new piece of memory contains the result.
-----------------*/
{
  BUF abuf ; char *p ;
  int c, diff, case_cmp ;

    buf_zero(&abuf) ;
    /* Comparison is case-insensitive if tag starts by '<' */
    case_cmp = *tag != '<' ; /* Case insensitive only for <.. */

    while((c = get_char()) != EOF) {
    	buf_achar(&abuf, c) ;
	if (c != *tag) continue ;
	abuf.pos = abuf.used - 1 ;
	for (diff = 0, p=tag+1; (diff == 0) && *p; p++) {
	    c = get_char() ;
	    if (c != EOF) buf_achar(&abuf, c) ;
	    if (!isalpha(c))   diff = *p - c ;
	    else if (case_cmp) diff = *p - c ;
	    else diff = toupper(*p) - toupper(c) ;
	}
	if (*p) {	/* No match... go on */
	    abuf.pos = 0 ;
	    continue ;
	}
	/* Match !! Remove the end_tag from buffer ... */
	abuf.text[abuf.used=abuf.pos] = 0 ;
    	break ;
    }
    return(abuf.text) ;
}
#endif

static char *null_fct() { return (""); }
#if 0	/* Not used... */
static char *setmode(char *new_mode) {
    if (!new_mode) new_mode = "" ;
    if (!*new_mode) output_mode = MODE_verb ;
    else output_mode = atoi(new_mode) ;
    return("") ;
}
#endif

static char *uppercase(char *str)  { return (trstr(str, MODE_upper)) ; }
static char *lowercase(char *str)  { return (trstr(str, MODE_lower)) ; }
static char *nl2cr(char *str)      { return (trstr(str, MODE_nl2cr)) ; }
static char *basename(char *filename)
{
  char *a, *p ;
    a = (char *)0 ;
    if (!filename) filename = a = get_string() ;
    for (p = filename + strlen(filename) ;
	(p > filename) && (p[-1] != '/'); p-- )  ;
    if (a) free(a) ;
    return(p) ;
}
static char *dirname(char *filename)
{
  char *s, *a, *p; static char name[128];
    a = (char *)0 ;
    if (!filename) filename = a = get_string() ;
    /* Look first if the sequence "/./" can be found... */
    for (p=filename; *p; p++) {
	if (*p != '/') continue ;
	if (p[1] != '.') continue ;
	++p;
	if (p[1] == '/') { p += 2; break ; }
    }
    if (!*p) p = filename ;
    s = name ;
    strncpy(s, p, sizeof(name)-1-(s-name)); name[sizeof(name)-1] = 0;
    for (p=name + strlen(name); (p>name) && (*p != '/'); p--) ;
    *p = 0;
#if DEBUG
    fprintf(_of, "\n----Dirname: Arel=%s, return %s\n", Arel ? Arel:"(nil)",
	name) ;
#endif
    if (a) free(a) ;
    if (p == name) return(".") ;
    return(name) ;
}
static char *ed_bibjvp(char *jvp) {	/* Edit 14-byte Jnl Vol Page Category */
  char *p, *a, categ; static char bibcode[40] ; int i, j ;

    a = (char *)0 ;
    categ = '.' ;
    if (!jvp) jvp = a = get_string() ;
    p = jvp; i = 0 ;
    while (isspace(*p)) p++ ; j = i + 5 ;
    while ((i<15) && (isGraph(*p))) bibcode[i++] = *(p++) ;
    while (i<j) bibcode[i++] = '.' ;
    while (isspace(*p)) p++ ; 
    for (j=0; isGraph(p[j]); j++); j = 4 - j;
    while (--j >= 0) bibcode[i++] = '.' ;
    while ((i<15) && (isGraph(*p))) bibcode[i++] = *(p++) ;
    while (isspace(*p)) p++ ;
    if (isalpha(*p)) categ = *(p++) ;
    bibcode[i++] = categ ;
    /* bibcode[i++] = *(p++) ; else bibcode[i++] = '.' ; */
    while (isspace(*p)) p++ ;
    /* V3.0: Page starts by 0 ==> remove the 0's */
    while (*p == '0') p++; if (*p == ' ') --p;
    j = 4 - strloc(p, ' ');
    while (--j >= 0) bibcode[i++] = '.' ;
    while ((i<15) && (isGraph(*p))) bibcode[i++] = *(p++) ;
    bibcode[i] = 0 ;
    /* Look in the category */
    while (isspace(*p)) p++ ;
    if (*p) {
	if (strncasecmp(p, "letter", 6) == 0) categ = 'L' ;
	else if (strncasecmp(p, "paper", 5) == 0) ;
	else fprintf(stderr, "++++Category for bibcode '%s' = %s ? \n",
	  bibcode, p) ;
    }
    bibcode[9] = categ ;
    if (a) free(a) ;
    return(bibcode) ;
}
static char *typein(char *msg)
{
  char answer[BUFSIZ] ;
    if (isatty(0) && (isatty(1)|isatty(2))) ;
    else {
	fprintf(_of, "\n\\verb\177**\typein can't work, not a terminal \177\n") ;
	return("\377\377") ;
    }
    fflush(stdout) ;
    fprintf(isatty(1) ? stdout : stderr, "\n\\typein{%s}: ", msg) ;
    fgets(answer, sizeof(answer), stdin) ;
    return(strdup(answer)) ;
}
static char *typeout(char *text)
{
    fprintf(stderr, "%s\n", text ? text : "${.typeout (nil)}") ;
    return("") ;
}
static char *reset(char *text)
{
    if (text && isdigit(*text)) lineno = atoi(text) ;
    else lineno = 1 ;
    return("") ;
}

static char *exec_notag() { output_mode |= MODE_notag; return("") ; }
static char *obeyline(char *txt) {	/* Write e.g. # at lines */
    output_mode |= MODE_obey ;
    if (txt && (*txt))
         begline = txt ;
    else begline = (char *)0 ;
    return("") ;
}
static char *exec_ignore(char *no)	/* Ignore Input */
/*++++++++++++++++
.PURPOSE  Execute the ${.ignore} ${.ignore 0} ${.ignore <}
-----------------*/
{
    output_mode = 0 ;
    if (no) while (*no) {
	if ((*no == '0') || (toupper(*no) == 'N'))
	    output_mode |= 1 ;
	else if (*no == '<') input_mode |= MODE_notag ;
	no++ ;
    }
    return("") ;
}
static char *exec_mark(char *arg)	/* Define a Marker display_depth */
{
    if (!arg) arg = "" ;
    /* if (!*arg) arg = tagname(current_tag) ; */
    mark_set(arg, display_depth) ;
    return("") ;
}
static char *exec_exit(char *arg)	/* Exit at Specified Marker	*/
{
  int m ; MARKER *p ; DIC *d ;
    if (!arg) arg = "" ;
    /* if (!*arg) arg = tagname(current_tag) ; */
    if ((p = mark_find(arg))) m = display_till = p->value ;
    else m = 0 ;
    if (display_depth > m) display_stop = -1 ;
    /* Remove the definitions not used */
    if (p && top_dic) while (p->top_dic != top_dic) {
	if (!top_dic) {
	    fprintf(stderr, "****exec_exit, no DIC any more ?!?\n") ;
	    break ;
	}
	d = top_dic->prev ;
	t1kill(top_dic) ;
	top_dic = d ;
    }
    /* Remove the empty marker(s) */
    if (!*arg) mark_kill(arg) ;
    return("") ;
}
static char *exec_save(char *arg)	/* Set collector to named buffer */
{
  char **a, *name, *input ;
    if (verbop&2) {
        printf("....exec_save(%p=", arg); fflush(stdout);
	if (arg) printf("'%s'", arg); 
	printf(")\n"); fflush(stdout);
    }
    a = getargs(arg, 2, 1) ;
    name = a[0] ;
    if (!name) name = "" ;
    if (!*name) name = tagname(current_tag) ;
    input = a[1] ;
    if (input) var_set(name, input) ;
    else {
	collector = var_set(name, "") ;
        output_mode = MODE_buf ;
    }
    if (a[0]) free(a[0]) ;
    if (a[1]) free(a[1]) ;
    return("") ;
}
static char *exec_squeeze(char *arg) 	/* Remove leading/trailing 	 */
{
  BUF *b ; char *s, *t, c ;
    if (!arg) arg = "" ;
    if (!*arg) arg = tagname(current_tag) ;
    b = var_find(arg) ;
    if (!b) return("") ;
    s = t = b->text; c = ' ';
    if (!s) return("") ;
    while(*t) {
        if (isspace(*t) && isspace(c)) ;
        else *(s++) = isspace(*t) ? ' ' : *t ;
        c = *(t++) ;
    }
    if (isspace(c) && (t>b->text)) --s ;
    *s = 0 ;
    return("") ;
}
static char *exec_trim(char *arg) 	/* Remove leading/trailing 	 */
{
  BUF *b ; char *s, *t ;
    if (!arg) arg = "" ;
    if (!*arg) arg = tagname(current_tag) ;
    b = var_find(arg) ;
    if (!b) return("") ;
    s = t = b->text;
    if (!t) return("") ;
    while (isspace(*t)) t++ ;		/* Leading blanks ...   	*/
    if (s != t) strcpy(s, t) ;
    s += strlen(s) ;
    while (--s >= b->text) {
	if (isspace(*s)) *s = 0 ;
	else break ;
    }
    return("") ;
}
static char *exec_fitsDISP(char *arg)		/* Takes datatype etc */
/* NOTE: Variable-length ==> lowercase format */
{
  static char datatype[] = "cusilfdb";
  static char foformat[] = "AIIIIFFL";		/* Corresponding fortran */
  static char buf[40] ;
  char *p, *b, *a;
  int dim, d, w;

    /* Which datatype ? */
    if ((p = sgml_attribute(arg, "datatype"))) 
	 d = strloc(datatype, tolower(*p));
    else d = strlen(datatype)-1;
    /* Width and arraysize */
    if ((a = sgml_attribute(arg, "arraysize"))) 
	 dim = atoi(a);
    else dim = 1;
    if ((p = sgml_attribute(arg, "width"))) 
	 w = atoi(p);
    else w = 0;
    if ((foformat[d] == 'A') && (w == 0)) { w = dim; dim = 0; }

    /* Repetition factor */
    b = buf;
    if (dim>1) { sprintf(b, "%d", dim); b += strlen(b); }

    /* format -- in lowercase for variable dimension! */
    *b = foformat[d]; 
    if (a) {
	while (isdigit(*a)) a++;
	if (*a == '*') *b = tolower(*b);
    }

    /* Width + precision */
    if ((p = sgml_attribute(arg, "precision"))) {
        if (isalpha(*p)) { 
	    *b = datatype[d] == 'd' ? 'D' : toupper(*p);
	    p++;
	}
    }
    if (*b) b++;
    *b = 0;
    if (w>0) { sprintf(b, "%d", w); b += strlen(b); }
    if (p && *p) {  /* Add the precision */
	sprintf(b, ".%s", p);
    }

    return(buf) ;
}
static char *exec_colspan(char *arg)		/* Insert the \t */
/* Argument: "1" for End -- other is type of alignment */
{
  static char buf[40] ;
  char *p; int n;
    p = sgml_attribute((char *)0, "colspan") ;
    if (!p) return("") ;
    n = atoi(p) ;
    if (n < 2) return("") ;

    if (arg) {
	if (isalpha(*arg)) {			/* multicol */
	    sprintf(buf, "\\multicolumn{%d}{%s}{", n, arg) ;
	    return(buf) ;
	}
	else if (*arg == '/') return("}") ;	/* End of multicol */
    }
    /* Just put the required number of tabs */
    if (n > sizeof(buf)) n = sizeof(buf);
    buf[--n] = 0;
    while (--n >= 0) buf[n] = '\t' ;
    return(buf) ;
}

static char *exec_append(char *arg)	/* Set collector to named buffer */
{
  char **a, *name, *input ;
    a = getargs(arg, 2, 1) ;
    name = a[0] ;
    if (!name) name = "" ;
    if (!*name) name = tagname(current_tag) ;
    input = a[1] ;
    if (input) var_append(name, input) ;
    else {
	collector = var_append(arg, "") ;
        output_mode = MODE_buf ;
    }
    if (a[0]) free(a[0]) ;
    if (a[1]) free(a[1]) ;
    return("") ;
}
static char *exec_clear(char *arg)	/* Clear a Buffer    */
{
  BUF *b ;
    if (!arg) arg = "" ;
    if (!*arg) arg = tagname(current_tag) ;
    if ((b = var_find(arg))) buf_clear(b) ;
    else var_set(arg, "") ;
    return("") ;
}
static char *exec_kill(char *arg)	/* Remove a variable */
{
    if (!arg) arg = "" ;
    if (!*arg) arg = tagname(current_tag) ;
    var_kill(arg) ;
    return("") ;
}
static char *gobble_spaces(char *arg)	/* Skip the following spaces */
{
    while (isspace(next_char()))
	get_char();
    return(arg);
}
static char *sub_entity(char *arg)
{
    return(tr_entity(arg)) ;
}

#if ALLOW_IF
static char *ifnum(char *arg) 	/* Compare */
{
  double v1, v2 ;
  int cond, op ;
  char *p ;
    for(p = arg; isspace(*p); p++) ;
    v1 = atof(p) ;
    if (issign(*p)) p++;
    while(isdigit(*p)) p++ ;
    if (*p == '.') for (++p; isdigit(*p); p++) ;
    while(isspace(*p)) p++ ;
    op = *(p++) ;
    v2 = atof(p) ;
    switch(op) {
      case '=': cond = v1 == v2 ;  break ;
      case '<': cond = v1 <  v2 ;  break ;
      case '>': cond = v1 >  v2 ;  break ;
      default:
      	fprintf(_of, "\n\\verb\177**Bad comparison %s\177\n", arg) ;
      	return("") ;
    }
    match_if(cond) ;
    return("") ;
}
static char *ifmatch(pattern, string)	/* If string Matches Pattern */
  char *pattern, *string ;
{
    match_if(strmatch(pattern, string)) ;
    return("") ;
}
static char *ifregexp(regexp, string)	/* If string Matches RegExpr */
  char *regexp, *string ;
{
  char buf[BUFSIZ] ; char ending_char;
    if (*regexp == '/') ending_char = *(regexp++) ;
    else ending_char = 0 ;
    compile(regexp, buf, buf+sizeof(buf), ending_char);
    match_if(strexp(buf, string)) ;
    return("") ;
}
#endif

static PTR_FCT fct_stack[MAX_FCT] = {
    /* #00 */	null_fct,
    /* #01 */	uppercase ,		/* ${.upper}	*/
    /* #02 */	lowercase ,		/* ${.lower}	*/
    /* #03 */	typein ,		/* ${.typein}	*/
    /* #04 */	basename ,		/* ${.basename}	*/
    /* #05 */	dirname ,		/* ${.dirname}	*/
    /* #06 */	nl2cr,			/* ${.nl2cr}	*/
    /* #07 */	exec_notag ,		/* ${.notag}	*/
    /* #08 */	exec_ignore ,		/* ${.ignore}	*/
    /* #09 */	include_file ,		/* ${.include}	*/
    /* #10 */	def_tag ,		/* ${.def}	*/
    /* #11 */	print_tags ,		/* ${.list}	*/
    /* #12 */	display_input ,		/* ${...}	*/
    /* #13 */	ed_bibjvp ,		/* ${.bibjvp}	*/
    /* #14 */	exec_save ,		/* ${.save}	*/
    /* #15 */	exec_append ,		/* ${.append}	*/
    /* #16 */	exec_kill ,		/* ${.kill}	*/
    /* #17 */	exec_clear ,		/* ${.clear}	*/
    /* #18 */	exec_sub ,		/* ${.sub}	*/
    /* #19 */	exec_gsub ,		/* ${.gsub}	*/
    /* #20 */	exec_mark ,		/* ${.mark}	*/
    /* #21 */	exec_exit ,		/* ${.exit}	*/
    /* #22 */	typeout,   		/* ${.typeout}	*/
    /* #23 */	reset ,    		/* ${.reset}	*/
    /* #24 */	obeyline ,    		/* ${.obeyline}	*/
    /* #25 */	exec_squeeze , 		/* ${.squeeze}	*/
    /* #26 */	exec_trim ,		/* ${.trim}	*/
    /* #27 */	exec_fitsDISP ,		/* ${.fitsDISP}	*/
    /* #28 */	exec_inject ,		/* ${.inject}	*/
    /* #29 */	exec_colspan ,		/* ${.colspan}	*/
    /* #30 */	gobble_spaces,		/* ${.gobble}	*/
    /* #31 */	sub_entity,		/* ${.entity}	*/
};
static int nfct_stack ;

PTR_FCT sgml_comment(PTR_FCT fct)
/*++++++++++++++++
.PURPOSE  Link a user-supplied function to a Symbol, form ${.xxx}
.RETURNS  The function previously defined.
.REMARKS  fct(char *comment_text) is called. fct is NULL for default action
		(ignore the comments)
-----------------*/
{
  PTR_FCT old_fct ;
    old_fct = save_comment ;
    save_comment = fct ;
    return(old_fct) ;
}

PTR_FCT sgml_fct(char *symbol, PTR_FCT fct)
/*++++++++++++++++
.PURPOSE  Link a user-supplied function to a Symbol, form ${.xxx}
.RETURNS  The previous definition of the function -- PTR_FCT(-1) for error
.REMARKS  symbol must be ${.name}
-----------------*/
{
  char *buf, *p; int len ;
  PTR_FCT old_symbol ;

    if (nfct_stack >= (MAX_FCT-1)) {
	putstring("\n\\verb\177**Can't add function '"); putstring(symbol);
	putstring("', array full**** \177\n") ;
    	return((PTR_FCT)(-1));
    }
    if (strncmp(symbol, "${." /*}*/, 3)) {
    	fprintf(_of,
    	  "\\verb\177****Symbol '%s' in sgml_fct must be of type ${.xxx}\177\n",
    	    symbol) ;
     	return((PTR_FCT)(-1)) ;
    }

    /* Find out the current definition of symbol */
    old_symbol = (PTR_FCT)0 ;
    if ((p = lookup(top_dic, symbol))) {
        if ((p = strchr(p, '#'))) {
	    while (*p && (!isdigit(*p))) p++ ;
	    old_symbol = fct_stack[atoi(p)] ;
	}
    }

    /* Allocate a buffer to hold the definition, and insert as a dic */
    len = strlen(symbol) + 7 ;
    buf = p = malloc(1+(len|7));
    strcpy(p, symbol); p += strlen(p);
    *(p++) = '#';
    strcpy(p, i2a(nfct_stack)) ; p += strlen(p) ;
    fct_stack[nfct_stack++] = fct;
    *(p++) = '\n';
    *p = 0 ;
    load_dic(buf, 0) ;
    return (old_symbol);
}

char *sgml_attribute (char *tag_name, char *attr_name)
/*++++++++++++++++
.PURPOSE  Retrieve the value of an attribute (in a static buffer)
.RETURNS  The equivalent / NULL when not found
.REMARKS  tagID is NULL ==> current tag
	tag_name must start by '='
-----------------*/
{
  static BUF mybuf ;
  char *tag, *s, *t, *p, aname[256] ;
  BUF *b ;
  int k ;
    if (tag_name && *tag_name) {
	b = var_find(tag_name) ;
	if (!b) {
	    strncpy(aname+1, tag_name, sizeof(aname)-2) ;
	    aname[sizeof(aname)-1] = 0 ;
	    aname[0] = '=' ;
	    b = var_find(aname);
	}
	if (!b) return((char *)0) ;
	tag = b->text ;
    }
    else tag = current_tag ;

    t = (char *)0 ;	/* Found tag with the explanation */
    s = attr_name ;	/* Name to look at */
    k = strarg(tag, s) ;
    if (tag[k]) t = tag;
    while(tag && (!t)) {
	k = strarg(tag, "ref") ;
	if (!tag[k]) break ;
	p = tag + k + 3;
	if (*(p++) != '=') break ;
	if (isquote(*p)) { k = strquote(p); p++; k-- ; }
	else for (k=0; isGraph(p[k]) && (p[k] != '>'); k++) ;
	/* Remember: name of IDs stored as '=' followed by id */
	if (k > sizeof(aname)-2) k = sizeof(aname)-2;
	strncpy(aname+1, p, k) ;
	aname[sizeof(aname)-1] = 0 ;
	aname[0] = '=' ;
	if ((b = var_find(aname))) {
	    tag = b->text ;
    	    k = strarg(tag, s) ;
	    if (tag[k]) t = tag;
	}
	else tag = (char *)0 ;
    }
    if (t) {		/* Found the explanation ! */
    	p = t + k ;
	while (isGraph(*p) && (*p != '=')) p++ ;
	if (*p != '=') {	/* No value specified... assume 1 */
	    return("1") ;
	}
	p++ ;		/* Skip the = sign */
	if (isquote(*p)) { k = strquote(p); p++; k-- ; }
	else for (k=0; isGraph(p[k]) && (p[k] != '>'); k++) ;
	buf_clear(&mybuf) ;
	buf_append(&mybuf, p, k) ;
    	return(mybuf.text) ;
    }
    return((char *)0) ;
}

static void exec_subshell(char *command, int len)
/*++++++++++++++++
.PURPOSE  Execute the command, and display the resut.
.RETURNS  ---
-----------------*/
{
  char x ;
#if POPEN
  FILE *fp ;
  char buf[BUFSIZ] ;
    x = command[len] ; command[len] = 0 ;
    fp = popen(command, "r") ;
    if (!fp) {
	fprintf(stderr, "****(%s)", inloc()) ;
	perror(command) ;
	command[len] = x ;
	return ;
    }
    while (fgets(buf, sizeof(buf), fp))
	putstring(buf) ;
    fclose(fp) ;
    command[len] = x ;
#else
    x = command[len] ; command[len] = 0 ;
    fprintf(_of, "****\\verb\177Can't launch subshell \"%s\"\177\n",
        command) ;
    command[len] = x ;
#endif
}

static void exec_symbol(char *symbol, int len)
/*++++++++++++++++
.PURPOSE  Execute the symbol of type ${...}
.RETURNS  ---
.REMARKS  symbol of form ${.xxx (arguments)}
-----------------*/
{
  char bufarg[80], *p, *e, *arg;
  int i, n, do_rescan ;
  PTR_FCT f ;

    if (verbop&2) printf("\n----exec_symbol('%s', len=%d) output_mode=%X\n",
        symbol, len, output_mode) ;
    if (len < 0) len = strlen(symbol) ;
    e = symbol + len ;
    i = strloc(symbol, ' ') ;
    if (i < len) {		/* There are arguments... */
    	arg = symbol + i ;
    	*arg = /*{*/ '}' ;
    }
    else arg = (char *)0 ;
    p = lookup(top_dic, symbol) ;
    if (arg) *arg = ' ' ;	/* Reset the blank	*/
    else arg = e ;
    if (!p) {
    	i = *e ; *e = 0 ;
        fprintf(_of, "\\verb\177****Undefined symbol: %s\177\n", symbol) ;
        *e = i ;
        return ;
    }
    i = strloc(p, '#') ;
    if (!p[i]) {
        fprintf(_of,
           "\\verb\177****Missing # in symbol definition: %s\177\n", p) ;
	return ;
    }

    /* Ignore leading blanks in arguments */
    while ((arg<e) && isspace(*arg)) arg++ ;

    /* Examine the Function: it may be required to compile regular
       expression if # followed by /
    */
    if (verbop&2) printf("----exec_symbol: ==> '%s'; arg='%s'\n", p,
        arg<e ? arg : "") ;
    do_rescan = 1 ;
    for (++i; ispunct(p[i]); i++) switch(p[i]) {
      case '/':			/* Compile the Regular Expression */
        			/* Expression already compiled ?. */
        if (*arg != '/') continue;
        n = compile_expr(arg) -1; 	/* Keep the ending /	  */
        if (verbop&2) {
            printf("    compiled %d bytes ---> arg='%s' nreg-1=%d...",
               n, arg, 1/*nreg*/-1) ;
            oterm = getch() ; putchar('\n') ;
            if (oterm == 3) exit(1) ;
    	}
        /* if (arg[n]) n++ ;	-- How many bytes used for Reg-Exp */
        arg[0] = '\05';
        /* arg[1] = nreg-1 ; */
        while (--n >= 1 /*2*/) arg[n] = ' ' ;
        if (verbop & 2) {
            printf("    --------------------> arg='%s'...", arg) ;
            oterm = getch() ; putchar('\n') ;
            if (oterm == 3) exit(1) ;
        }
    	continue ;
      case '-':	/* NO RESCAN */
        do_rescan = 0 ;
        continue ;
      default:
        if (!isdigit(p[i])) fprintf(_of,
            "****Bad Symbol Definition: '%s'\n", p) ;
        return ;
    }
    if (verbop&2) printf("                 p[i]='%s'", p+i) ;
    i = atoi(p+i) ;
    if (verbop&2) printf(" ==> fct #%d = %p\n", i, fct_stack[i]) ;
    f = fct_stack[i] ;
    if (!f) {
        fprintf(_of, "\\verb\177****Inexistant symbol # in: %s\177\n", p) ;
	return ;
    }

    /* Ignore trailing blanks in arguments */
    while ((arg < e) && isspace(e[-1])) e-- ;
    if (arg < e) { 	/* There are arguments... */
	i = (e-arg);
	if (i < (sizeof(bufarg)-1)) {
	    p = (char *)0 ;
	    strncpy(bufarg, arg, i) ;
	    arg = bufarg ;
	}
	else {
	    p = malloc(i+1) ;
	    strncpy(p, arg, i) ;
	    arg = p ;
	}
	arg[i] = 0 ;
	/* The arguments have generally to be rescanned */
	if (do_rescan && ((output_mode&MODE_verb) == 0)
	   && strchr(arg, '$')) {
	    e = rescan(arg, -1) ;
	    if (p) free(p) ;
	    p = arg = e ;
	}
    }
    else p = arg = (char *)0 ;
    e = (*f)(arg) ;
    if (p) free(p) ;
    if (e && *e) PutString(e) ;
}

/*==================================================================
		Matching Routines
 *==================================================================*/

static char *rescan(char *text, int len)
/*++++++++++++++++
.PURPOSE  Scan a text, collect the result
.RETURNS  The equivalence in a newly allocated buffer.
-----------------*/
{
  BUF *old_collector, *old_inbuf ; char *old_begline ;
  int old_mode, old_input;
  BUF bin, but ;
  char *p, x ;

    if (verbop&2) {
        printf("----Rescan(\"%s\", len=%d)...", text, len) ;
        oterm = getch() ; putchar('\n') ;
        if (oterm == 3) exit(1) ;
    }
    if (len < 0) len = strlen(text) ;
    buf_zero(&bin) ; buf_zero(&but) ;

    /* Save the Current Environment, and install the string as Input */
    old_mode = output_mode ; output_mode = 2 ;
    old_input = input_mode ; input_mode |= MODE_rescan ;
    old_begline = begline ; begline = (char *)0 ;
    old_collector = collector ; collector = &but ;
    old_inbuf = inbuf ;  inbuf = &bin;
    inbuf->size = inbuf->used = len ;
    inbuf->text = text ;
    x = text[len] ; text[len] = 0 ;

    /* Execute the text -- a shortcut for ${...} (getinput) is used */
    if (strcmp(text, "${...}") == 0) {
	inbuf->pos = inbuf->used ;
	input_mode &= ~MODE_rescan ;
    }
    display_input((char *)0) ;
    p = collector->text ;
    if (!p) p = "" ;

    /* Restore the environment */
    text[len] = x ;
    inbuf = old_inbuf ;
    collector = old_collector ;
    begline = old_begline ;
    input_mode = old_input ;
    output_mode = old_mode ;

    return(strdup(p)) ;
}

static int match_entity()
/*++++++++++++++++
.PURPOSE  Triggered after a '&' is read
.RETURNS  EOF if end (Last byte read).
-----------------*/
{
  char buf[30], *p ;
  int i, c ;
    i = 0 ;
    buf[i++] = '&' ;
    while ((c = next_char()) != EOF) {
	if ((i >= (sizeof(buf)-2)) || isspace(c) || (c == '<')) {
	    buf[i] = 0 ;
	    PutString("&[++++Missing semi-colon in entity]");
	    PutString(buf+1);
	    return(c) ;
	}
	buf[i++] = get_char();
	if (c == ';') break ;
    }
    buf[i] = 0 ;

    /* Can it be a simple number (Latin) ? */
    if (buf[1] == '#') {
	i = atoi(buf+2) ;
	if ((i&(~127)) == 128) {	/* Range  128..255 */
	    PutChar(i) ;
	    return(c) ;
	}
	/* return(c) ; */
    }

    /* Get the Translation */
    if ((p = lookup(top_dic, buf))) {
    	p += strloc(p, '=');
    	if (*p) p++ ; 			/* Skip the = sign */
        PutString(p) ;
        return(c) ;
    }

    /* Don't know what this is -- take defaults according to tex/tex2 */
    i = strloc(buf, ';') ; buf[i+1] = 0 ;
    if (otex&3) buf[0] = '{', buf[i] = '}' ;
    else buf[0] = '[', buf[i] = ']' ;
    if ((otex&3) == 1) PutChar(buf[0]), PutChar('\\'), PutString(buf+1) ;
    else PutString(buf) ;
    return(c) ;
}

static int match_symbol(char *symbol)
/*++++++++++++++++
.PURPOSE  Triggered to exec what's required in ${.. .  }
.RETURNS  Length of processed part of string
.REMARKS  symbol starts by ${ and symbol[len] = '}'
	If symbol is NULL, assume $ read, get from input
-----------------*/
{
  static BUF abuf ;
  char aname[256] ;
  BUF *b ;
  char *p, *e, *s, *t, *tag, a ;
  int len, k;

    /* Input not specified: get it from Input */
    if (!symbol) {
	buf_clear(&abuf) ;
	buf_achar(&abuf, '$') ;
	k = get_char() ;	/* Get the opening brace */
	buf_achar(&abuf, k) ;
	get_brace(&abuf) ;
	symbol = abuf.text ;
    }
    len = strbrace(symbol+2) + 2 ;
    if (!symbol[len]) {	/* No matching brace...	*/
    	PutString("++++(no matching brace in symbol)") ;
    	PutString(symbol) ;
    	PutString("++++\n") ;
    	return(len) ;
    }

    if (len == 2) {	/* CURRENT TAG NAME	*/
	PutString(tagname((char *)0)) ;
	return(len+1) ;
    }
    a = 0 ;		/* Set to '@' for Attr, '$' for variab.	*/
    switch(symbol[2]) {
      case '*':		/* CURRENT FULL TAG	*/
	PutString(fulltag((char *)0)) ;
	return(len+1) ;
      case '!' :	/* Execute a SubShell	*/
	exec_subshell(symbol+3, len-3) ;
	return(len+1) ;
      case '.' :	/* Execute Specified Action	*/
    	/* Warning: len smaller than actual symbol length
    	   to get correctly the arguments to the symbol	*/
    	exec_symbol(symbol, len) ;
    	return(len+1) ;
      case '$':		/* Accept $ or $$ alone */
	for (k=2; (k<len) && (symbol[k] == '$'); k++) ;
	if (k < len) break;
	PutText(symbol+2, len-2) ;
	return(len+1) ;
      case '<': case '>': case ' ':
	PutText(symbol+2, len-2) ;
	return(len+1) ;
      case '^' :	/* Beginning of Line	*/
	PutChar('\r') ;
	if (len > 3) PutText(symbol+3, len-3) ;
    	return(len+1) ;
      case '\\' : 	/* Issue the Octal Char	*/
	output_mode |= MODE_force ;
	PutChar(thebyte(symbol+2)) ;
	output_mode &= ~MODE_force ;
    	return(len+1) ;
      case '%':		/* Edit Formatted Text  */
	a = symbol[len]; symbol[len] = 0;
	exec_fmt(symbol+2);
	symbol[len] = a;
	return(len+1);
    }

    /* The name of the symbol, when starting by
       @ => must be an attribute of the current tag
       $ => is the name of a variable
       We look here in the current tag, and in the referenced ones
    */
    tag = current_tag;
    t = (char *)0 ;	/* Found tag with the explanation */
    s = symbol+2 ;	/* Name to look at */
    if ((*s == '$') || (*s == '@')) a = *(s++) ;
    k = strarg(tag, s) ;
    if (tag[k]) t = tag;
    if (a == '$') t = (char *)0 ;
    while(tag && (!t)) {
	k = strarg(tag, "ref") ;
	if (!tag[k]) break ;
	p = tag + k + 3;
	if (*(p++) != '=') break ;
	if (isquote(*p)) { k = strquote(p); p++; k-- ; }
	else for (k=0; isGraph(p[k]) && (p[k] != '>'); k++) ;
	/* Remember: name of IDs stored as '=' followed by id */
	if (k > sizeof(aname)-2) k = sizeof(aname)-2;
	strncpy(aname+1, p, k) ;
	aname[sizeof(aname)-1] = 0 ;
	aname[0] = '=' ;
	if ((b = var_find(aname))) {
	    tag = b->text ;
    	    k = strarg(tag, s) ;
	    if (tag[k]) t = tag;
	}
	else tag = (char *)0 ;
    }
    if (t) {		/* Found the explanation ! */
    	p = t + k ;
	while (isGraph(*p) && (*p != '=')) p++ ;
	while (isspace(*p)) p++; 	/* V2.91: accept attribute = value */
	if (*p != '=') {	/* No value specified... assume 1 */
	    PutChar('1') ;
	    return(len+1) ;
	}
	p++ ;		/* Skip the = sign */
	while (isspace(*p)) p++; 	/* V2.91: accept attribute = value */
	if (isquote(*p)) { k = strquote(p); p++; k-- ; }
	else { k = strloc(p, ' ') ; if (p[k-1] == '>') k-- ; }
	PutTextUnquoted(p, k) ;
    	return(len+1) ;
    }
    if (a == '@') 	/* Only attribute -- if not, exit */
	return(len+1) ;

    /* ... or can be the name of a variable
       Accept $name$1-3 (substrings) -- NOT CODED !
    */
    e = s + strloc(s, /*{*/'}') ;
    if ((p = strchr(s, '$'))) {
	if (p < e) e = p ;
    }
    k = *e ; *e = 0 ;
    b = var_find(s) ;
    *e = k ;
    if (b) {		/* Here, one should CODE the substitude etc */
	if (b->text) PutStringUnquoted(b->text) ;
	return(len+1) ;
    }
    if (a == '$') 	/* Only Variable  -- if not, exit */
	return(len+1) ;

    /* Symbol can be defined in dictionary... */
    if ((p = lookup(top_dic, symbol))) {
    	p += strbrace(p) ;
    	if (*p) p++ ;
    	if (*p == ':') {	/* Have to rescan the :	*/
    	    p++ ;
    	    p = rescan(p, -1) ;
    	    PutString(p) ;
    	    free(p) ;
    	    return(len+1) ;
    	}
    	if (*p == '=') p++ ;
    	PutString(p) ;
    	return(len+1) ;
    }

    /* ... Or Finally from the Environment ... */
    for (p=e=symbol+2; isid(*e); e++) ;
    k = *e ; *e = 0 ;
    p = getenv(p) ;
    *e = k ;
    if (p) {
    	PutString(p) ;
    	return(len+1) ;
    }

    /* Can't understand... Just display message */
    if (osilent) return(len+1) ;

    k = symbol[len+1]; symbol[len+1] = 0 ;
    fprintf(stderr, "++++(%s) Can't understand '%s'", inloc(), symbol) ;
    fprintf(_of, "\n++++Can't understand '%s'", symbol) ;
    symbol[len+1] = k ;
    if (current_tag) fprintf(_of, " (within %s)", current_tag) ;
    fputc('\n', _of) ;
    return(len+1) ;
}

#if ALLOW_IF
static char *match_if(int cond)
/*++++++++++++++++
.PURPOSE  Match an ${.if } .. ${.else} .. ${.fi} Contition
.RETURNS  The text coresponding to the clause
.REMARKS  We assume that se start just after the ${.ifxx.}
	The matching clause (then or else) is saved.
-----------------*/
{
  static BUF bsym ;	/* To save symbols */
  BUF accu ;
  int state, depthif, c, i ;

    buf_zero(&accu) ;
    state = cond ? 1 : 0 ;
    depthif = 1 ;

    /* Accumulates the input into the 'accu' buffer	*/
    while (depthif > 0) {
    	c = get_char() ;
    	if (c == EOF) { depthif = 0 ; continue ; }
    	if (c != '$') {
    	    if (state) buf_achar(&accu, c) ;
    	    continue ;
        }
        c = get_char() ;
    	if (c != '{' /*}*/) {
    	    if (state) buf_achar(&accu, c) ;
    	    continue ;
        }
        buf_clear(&bsym) ;
        buf_append(&bsym, "${" /*}*/, 2) ;
        get_brace(&bsym) ;
        if (strcmp(bsym.text, "${.fi}") == 0) {
    	    depthif--;
    	    if (depthif == 0) continue ;	/* Gobble ${.fi} */
	}
    	else if ((strcmp(bsym.text, "${.else}") == 0)
    	     && (depthif == 1)) {
    	    state ^= 1 ;
    	    continue ;
    	}
    	else if (strncmp(bsym.text + 2, ".if", 3) == 0)
    	    depthif++ ;
    	if (state) buf_append(&accu, bsym.text, -1) ;
    }
    return(accu.text) ;
}
#endif

#if 0
static int ignore_until(char *text)
/*++++++++++++++++
.PURPOSE  Read until the text is read
.RETURNS  The last read byte -- normally the last char of text, or EOF
.REMARKS  When match_co
-----------------*/
{
  int c, i, bytes ;
    i = 0 ;
    while(1) {
	c = get_char() ;
	if (c == EOF) return(c) ;
	if (c != text[i]) { i = 0; continue; }
	if (!text[++i]) return(c) ;	/* We're at end of text = Success! */
    }
    return(EOF) ;
}
#else
static int match_comment()
/*++++++++++++++++
.PURPOSE  Read the comment, save it in comment_buf,
	   and execute the associated action
.RETURNS  0 (comment in "comment" buffer, nothing more to do)
	  Non-zero (last read byte)
.REMARKS  When save_comment is on, the function is called with the
		contents of the comment.
	Unless MODE_obey is on, the spaces following the comment are ignored.
-----------------*/
{
  static char end_comment[] = "-->" ;
  static BUF  bcom ;
  int c, i, old_mode, old_output ;
  char *p, *old_current_tag ;
  BUF *old_inbuf ;

    i = 0 ;
    if (!comment_buf) comment_buf = var_find("comment") ;
    if (!comment_buf) comment_buf = var_set("comment", "");
    buf_clear(comment_buf) ;
    while((c = get_char()) != EOF) {
	buf_achar(comment_buf, c) ;
	if (c != end_comment[i]) { i = 0; continue; }
	if (!end_comment[++i]) break;	/* We're at end of text = Success! */
    }
    /* Remove the ending --> */
    comment_buf->used -= i ;
    comment_buf->text[comment_buf->used] = 0 ;
    if (save_comment) { (*save_comment)(comment_buf->text) ; return(c); }

    /* V3.4: ignore the blanks after the comment -- except in MODE_obey */
    if (output_mode&MODE_obey);
    else gobble_spaces((char *)0);

    /* Is there something to do with the comment ? */
    if ((p = lookup(top_dic, "<!-->"))) {
	p += 5 ;
	if (*p) p++ ;
	if (!*p) p = (char *)0 ;
    }
    if (!p) return(c) ;	/* No action associated to Comment */

    /* Save current environment */
    old_mode = input_mode ; input_mode |= MODE_rescan ;
    old_output = output_mode ;
    old_current_tag = current_tag; current_tag = comment_tag ;
    old_inbuf = inbuf ;  inbuf = &bcom ;

    inbuf = &bcom ;
    bcom.text = p ;
    bcom.used = strlen(p) ;
    bcom.pos  = 0 ;
    display_input("") ;

    /* Restore curent environment */
    inbuf = old_inbuf ;
    input_mode = old_mode ;
    output_mode = old_output ;
    current_tag = old_current_tag ;

    return(c) ;
}
#endif

static int match_tag()
/*++++++++++++++++
.PURPOSE  Triggered after a '<' is read
.RETURNS  EOF = final
-----------------*/
{
  int old_mode; char *old_current_tag ; char *old_begline ;
  BUF abuf, bin, *old_inbuf, *b, *old_collector ;
  char tag[80], *p;
  int c, c2, i, closed, nbra, smode ;

    /* Save current environment */
    old_mode = output_mode ;
    old_current_tag = current_tag ;
    old_begline   = begline ;
    old_collector = collector ;

    /* Read the input, look for closing > */
    b = &abuf; buf_zero(b) ;
    i = nbra = 0 ;
    tag[i++] = '<' ;
    c2 = next_char() ;
    closed = 0 ;			/* Put closed='/' for no-close */
    while ((c = get_char()) != EOF) {
        if (b->text) buf_achar(b, c);
        else {
            tag[i++] = c ;
	    if ((c2 == '!') && (i >= 4) && (tag[2] == '-') && (tag[3] == '-')) {
		/* It's a comment ! */
		c = match_comment(); /* c = ignore_until("-->") ; */
		return(c) ;
	    }
            if (i >= sizeof(tag)-1) buf_append(b, tag, i);
        }
        if (isquote(c)) {	/* Match Quote */
            if (!b->text) buf_append(b, tag, i) ;
            get_until(b, c) ;
	    continue ;
        }
	if ((c == '[') && (c2 == '!')) {	/* Match <!DOCTYPE [ ... ]> */
	    nbra++ ;	/* Locate <![CDATA[ */
	    if ((nbra == 2) && (!b->text)) {
		 if (strncmp(tag, "<![CDATA", 8) == 0) {
		     c = next_char();
		     /* Accept CDATA followed  by a newline (V3.2) */
		     if (c == '\n') { c = get_char(); PutChar('\r'); }
		     output_mode |= MODE_verb ;
		     display_input("]]>") ;
		     c = 0 ;
		     goto EndMatchTag ;
		 }
	         buf_append(b, tag, i);
	    }
	    else if ((i>8) && (!b->text)) 	/* Match <!DOCTYPE[ ... */
	 	buf_append(b, tag, i);
	    if (b->text) get_brack(b) ;
	    continue ;
	}
        if (c == '>') {
	    if (c2 != '!') break ;
	    if ((p = b->text)) i = b->used ;
	    else p = tag ;
	    if ((strncmp(p, "<!--", 4) == 0) &&	/* Match Comment */
		(strncmp(p+i-3, "-->", 3) != 0)) continue ;
	    break ;
	}
	closed = c ;
    }
    if (closed != '/') closed = 0 ;
    if (c2 == '!') return(c) ;		/* A comment */
    if (b->text) {
	current_tag = b->text;
	/* If there is an ID for this tag, save it with its name */
	if (strstr(current_tag, " ID=")) save_tag(b->text) ;
    }
    else tag[i] = 0, current_tag = tag ;
    if (c != '>') {
    	fprintf(_of, "\\verb\177****Unclosed tag %s\n****(Skipped)\177\n",
    	    current_tag) ;
    	goto EndMatchTag ;
    }

    /* Get the Translation. */
    p = lookup_with_attr (top_dic, current_tag) ;
    if (!p) {
	if ((otex&3) == 1) p = alatex_tag ;
	else p = alasoft_tag ;
    	/* fprintf(_of, "****Unknown  tag %s\n****(Skipped)\n", current_tag); */
    }
    if (verbop&2) {
        printf("....match_tag\t===>%s ...", p) ;
        oterm = getch() ; putchar('\n') ;
        if (oterm == 3) exit(1) ;
    }

    /* Simple tag (begin/end not required): just display !	*/
    p += strloc(p, '>') ; if (*p) p++ ;

    switch(*p) {
      case ':':		/* No end tag required 	*/
    	/* NO BREAK */
      case '=':		/* Standard actions	*/
     	++p ;
    	while (*p) {
	    /* if (output_mode&MODE_verb) { PutString(p); break ; } */
	    if (p[0] == '<') {
	        if (output_mode & MODE_notag) {
		    p += strangle(p) ;
		    if (*p) p++ ;
		    continue ;
		}
                old_inbuf = inbuf ;  inbuf = &bin;
                smode = input_mode ; input_mode |= MODE_rescan ;
                inbuf->size = inbuf->used = strlen(p) ;
		inbuf->pos = 1 ;
                inbuf->text = p ;
                match_tag() ;
                p += inbuf->pos ;
		inbuf = old_inbuf ;
		input_mode = smode ;
		continue ;
	    }
    	    if (p[0] != '$') { PutChar(*p) ; p++ ; continue; }
    	    if (p[1] != '{') { PutChar(*p) ; p++ ; continue; } /*}*/
	    if (closed && (strncmp(p, "${...}", 6) == 0))  /* NO INPUT */
	         i = 6 ;
    	    else i = match_symbol(p);
            p += i ;
    	}
    	goto EndMatchTag ;
    }
    fprintf(_of, "****Ignored Equivalence of %s%s\n", current_tag, p) ;

    /* Restore current environment */
  EndMatchTag:
    output_mode = old_mode ;
    current_tag = old_current_tag ;
    begline   = old_begline ;
    collector = old_collector ;
    buf_free(b) ;
    return(c) ;
}

static char *display_input(char *end_tag)
/*++++++++++++++++
.PURPOSE  Display from current input file, until an end-of-tag (</) is found
.RETURNS  Pointer to "" / NULL if error.
.REMARKS  end_tag completely read. Default end_tag is </> from current_tag.
-----------------*/
{
  static BUF abuf ;	/* Buffer to save one tag */
  int c, mode;
  char etag[LEN_TAG+4], *p ;

    ++display_depth ;
    display_stop = 0 ;
    /* In -trim option:
	otrim&0x10 = At least one non-blank char already read.
	otrim&0x20 = Issue a BLANK before next non-blank character
    */
    if (otrim) otrim &= ~0x30;
    /* We're reading from the input file. However, when called from
      rescan, we've to take the input from the saved buffer,
      but a recursive call must be from the input.
    */
    if (input_mode&MODE_rescan) {		/* Direct call from rescan */
	/* if (input_mode&MODE_nofile); else old_inbuf = (BUF *)0 ; */
	input_mode -= MODE_rescan ;	/* Next time, use input	*/
    }
#if 0
    else {			/* Be sure file input  	*/
	old_inbuf = inbuf ;
	inbuf = (BUF *)0 ;
    }
#endif
    if (!end_tag) end_tag = "" ;
    /* Define the default end_tag */
    if ((*end_tag != '<') && (*end_tag != ']') && current_tag && *current_tag) {
        c = 0  ;
        etag[c++] = '<' ; etag[c++] = '/' ;
        for (p = current_tag+1;
	     isGraph(*p) && (*p != '>') && (c<sizeof(etag)-2); p++)
	    etag[c++] = *p ;
        etag[c++] = '>' ;
        etag[c] = 0 ;
        end_tag = etag ;
    }
    if (verbop&2) {
        printf("----Enter display_input #%d (%s)...", display_depth, end_tag) ;
        oterm = getch() ; putchar('\n') ;
        if (oterm == 3) exit(1) ;
    }
    while ((c = get_char()) != EOF) {
	if (otrim) {
	    if (isspace(c)) {
		if (otrim&0x10) otrim |= 0x20;
		continue;
	    }
	    else {
		if (otrim&0x20) { PutChar(' '); otrim &= ~0x20; }
		otrim |= 0x10;		/* Non-blank character read. */
	    }
	}
	switch(c) {
          case '&':
	    if (output_mode&MODE_verb) goto put_char ;
	    match_entity() ; continue ;
          case ']':		/* Look for End of Data ]]> */
	    if (!(output_mode&MODE_verb))
	        goto put_char ;
	    if (c != end_tag[0]) goto put_char ;
#if 0	    /* Modified V2.7 */
            c = next_char() ;
	    if (c != end_tag[1]) { c = end_tag[0]; goto put_char ; }
	    get_char() ;
	    c = next_char() ;
	    if (c != end_tag[2]) {
	        PutChar(end_tag[0]);
	        c =  end_tag[1];
	        goto put_char;
	    }
	    get_char() ;
#else	    /* New Code V2.7 */
 	    { int i, n;
	        for (n=1; end_tag[n]; n++) {
		    c = next_char();
		    if (c != end_tag[n]) break;
		    get_char();
	        }
	        if (end_tag[n]) {	/* Failed matching end_tag */
		    --n;		/* Last char was not read. */
		    for (i=0; i<n; i++) PutChar(end_tag[i]);
		    c = end_tag[n];
		    goto put_char;
	        }
	    }
#endif
	    /* The ]]> is MATCHED */
	    p = "" ;
	    goto Completed ;
          case '<':
            c = next_char() ;
            if ((c == '&') || (c == '%')) {	/* GLU */
                PutChar('<') ;
                c = get_char() ;
                PutChar(c) ;
                continue ;
            }
	    if (c == '?') {		/* XML Definition */
	        if (output_mode&MODE_verb) PutChar('<') ;
	        else {
	            buf_clear(&abuf) ;
		    buf_achar(&abuf, '<') ;
		    get_until(&abuf, '>') ;
		    p = abuf.text ;
		    p += strlen(p)-2 ;
		    if (*p != '?') fprintf(stderr,
		    "++++Line %d: XML-definition not closed: %s\n",
		        lineno, abuf.text) ;
	        }
	        continue ;
	    }
            if (c != '/') {
	        if (input_mode&MODE_notag) PutChar('<') ;
	        else if (output_mode&MODE_verb) PutChar('<') ;
	        else {
		    mode = input_mode ;
		    match_tag() ;
		    input_mode = mode ;
	        }
	        continue ;
	    }
            buf_clear(&abuf) ;
            buf_achar(&abuf, '<') ;
            get_until(&abuf, '>') ;
            if (stundiff(abuf.text, end_tag, abuf.used) == 0) {
	        p = "" ;
	        goto Completed ;
	    }
	    /* It's an unmatching closgin tag 	*/
	    if (output_mode&MODE_verb) {
	        PutString(abuf.text) ;
	        continue ;
	    }
            /* I just ignore spurious closing tags */
	    abuf.text[1] = '<' ;	/* Look definition: no ${...} ? */
	    if ((p = lookup(top_dic, &(abuf.text[1])))) {
	        p += strloc(p, '>') ;
	        if (*p) p++ ;
	        if (*p) p++ ;
	        while (*p) {
		    if (p[0] != '$') { p++; continue; }
		    if (strncmp(p, "${...}", 6) == 0) break ;
		    p++ ;
	        }
	    }
	    else p = "(non-defined tag)" ;
	    if (*p) {
	        abuf.text[0] = '<'; abuf.text[1] = '/' ;
	        fprintf(stderr,
		  "++++(%s) Found %s instead of %s, current_tag:%s\n",
		  inloc(), abuf.text, end_tag, current_tag) ;
	        if (err_text) putstring(err_text) ;
	    }
            continue ;
          case '$':
	    if (next_char() != '{' /*}*/) {
	        if (otex&1) PutString("\\$") ;
	        else goto put_char ;
	    }
	    if (output_mode&MODE_verb) PutChar(c) ;
	    else match_symbol((char *)0) ;
	    if (display_stop) {		/**** Was asked to stop !! */
	        fprintf(stderr, "++++(%s) Force closing %s, current_tag:%s\n",
		    inloc(), end_tag, current_tag) ;
	        p = (char *)0 ;
	        goto Completed ;
	    }
            continue ;
          case '\\':
            if ((otex&1) == 0) goto put_char ;
            PutString("{\b}") ;
            continue ;
          case '%': case '{': case '}': case '#': case '^': case '_':
	    if (output_mode&MODE_verb) goto put_char ;
	    if (otex&1) PutChar('\\') ;
	    goto put_char ;
          put_char:
          default:
            PutChar(c) ;
            continue ;
        }
    }
    p = (char *)0 ;
  Completed:
    /* if (old_inbuf) inbuf = old_inbuf ; */
    --display_depth ;
    if (verbop&2)
        printf("\n----Exit   display_input #%d (%s) ==> %s\n",
            display_depth, end_tag, p ? p : "(nil)") ;
    return(p) ;
}

/*==================================================================
		Executions
 *==================================================================*/
static void init()
/*++++++++++++++++
.PURPOSE  Just initalize the various pieces
.RETURNS  ---
-----------------*/
{
  static char regex[] = "/^$/";
  int i;

    if (regex[0] != '/') return ;	/* Compiled expression... */

    /* Install tagChar -V3.5) */
    for (i=0; i<256; i++)  { tagChar[i] = isGraph(i) ? 1 : 0 ; }
    tagChar['<'] = tagChar['>'] = tagChar['/'] = tagChar['\\'] = 0;
    tagChar['&'] = tagChar[';'] = 0;

    if (!_of) _of = stdout ;
    if (!theprogram) theprogram = "sgml2" ;
    compile_expr("/^$/") ;	/* NULL regular expression */
    regex[0] = 0;
    load_dic(default_defs, DIC_SORTED); 	/* Install the Service */
    if (otex&8) char8b = (char **)0 ;
    else if (otex&2) char8b = tex8soft ;
    else if (otex&3) char8b = tex8 ;
    else char8b = asc8 ;

    /* Install language-specific Definitions */
    if (otex&2) load_dic(tex2_tags, DIC_SORTED) ;
    else if (otex&1) load_dic(tex_tags, DIC_SORTED) ;
    else load_dic(asc_tags, DIC_SORTED) ;

    /* Count the standard Functions */
    for (nfct_stack=0; fct_stack[nfct_stack]; nfct_stack++) ;
}

static int do_conv(int opt)
/*++++++++++++++++
.PURPOSE  Execute the conversion on opened input file
.RETURNS  Status (EOF = end of file)
;REMARKS  The mime-type MUST start by a 'C'
------------------*/
{
  int status ;
  char *old_tag ;
  char buf[512];
    old_tag = current_tag ;
    current_tag = "" ;
    /* V2.9: Display the Mime Header -- until a blank line */
    if (omime && (next_char() == 'C')) { char *e;
	if (!_of) _of = stdout;
	while (fgets(buf, sizeof(buf), infile)) {
	    fprintf(_of, "%s", buf);
	    for (e=buf+strlen(buf)-1; (e>=buf) && isspace(*e); e--) ;
	    if (e<buf) break;
	}
	fflush(_of);
	omime = 0;
    }
    status = display_input("") ? 0 : EOF ;
    match_symbol("${^}") ;	/* Terminate line */
    current_tag = old_tag ;
    return(status) ;
}

char *sgml_result()
/*++++++++++++++++
.PURPOSE  Get the result, and reset the collector
.RETURNS  The string found (must be cloned !)
------------------*/
{
    if (!collector) return((char *)0) ;
    collector->used = 0 ;
    return(collector->text) ;
}

int sgml_exec(char *text)
/*++++++++++++++++
.PURPOSE  Execute the conversion on a String
.RETURNS  Status (EOF = end of file)
------------------*/
{
  static BUF local_buf ;
  BUF *old_inbuf ;
  int old_mode, status ;

    init() ;
    old_inbuf = inbuf ;
    old_mode = input_mode ;
    if (text) {
        local_buf.text = text ;
        local_buf.used = strlen(text);
        local_buf.pos  = 0 ;
	inbuf = &local_buf;
	input_mode = MODE_nofile ;
    }
    else if (!infile) infile = stdin ;
    status = do_conv(0) ;
    input_mode = old_mode ;
    inbuf = old_inbuf ;
    return(status) ;
}

char *sgml_conv(char *text)
/*++++++++++++++++
.PURPOSE  Execute the conversion on a String, and get the Result
.RETURNS  Status (EOF = end of file)
------------------*/
{
  static BUF local_buf ;
  BUF *old_collector ;
  int old_mode;

    old_collector = collector ; collector = &local_buf ;
    old_mode = output_mode ; output_mode = MODE_buf ;
    buf_clear(&local_buf) ;
    sgml_exec(text) ;
    output_mode = old_mode ;
    collector = old_collector ;
    return(local_buf.text) ;
}

int sgml_opt(char *text)
/*++++++++++++++++
.PURPOSE  Set the option
.RETURNS  0=ignored  1=OK  -1=Error
------------------*/
{
  int o;
  char *p;
    p = text;
    if (*p == '-') switch(p[1]) {
      case 'v':
	if (strncmp(p, "-ver", 4) == 0) {
	    printf("sgml2 (CDS) Version %s\n", VERSION);
	    exit(0);
	}
        verbop = 1 ;
        if (isdigit(p[2])) verbop = atoi(p+2) ;
        else if (p[2]) verbop |= 2 ;		/* Debugging */
        return(1);
      case 'h': printf("%s\n%s\n", usage, help) ;
        exit(0) ;
      case 's':
	osilent = 1 ;
	return(1);
      case 'k':		/* Keep new lines  */
	optnl = 1 ;
	output_mode |= MODE_obey ;
	return(1);
      case 'f':		/* Dic from a file */
	init() ;	/* ... required, otherwise overwritten ! */
        o = p[2] ;
	if (o == 0);
	else if (o == 'i') o = DIC_IGNORE;
	else if (o == 's') o = DIC_SORTED;
	else goto bad_option;
	while (isgraph(*p)) p++; while (isspace(*p));
	if (!*p) { p = *++Argv; --Argc; }
        p = load_file(p, 0) ;
        if (p) load_dic(p, o) ;	/* Option 's' if file sorted */
        return(1);
      case 'a':		/* -a8 option	*/
	if (p[2] == '8') otex |= 8 ;
	else otex = 0 ;
	return(1);
      case 't':		/* TeX option	*/
	if (p[2] == 'r') { otrim = 1; return(1); }
	otex = 1; p++ ;
	while(isalpha(*p)) p++ ;
	if (isdigit(*p)) otex = *p - '0' ;
	return(1);
      case 'e':		/* Error Text	 */
	for (p+=2; isspace(*p); p++) ;
	if (*p) err_text = p;
	else { err_text = *++Argv; Argc--; }
	return(1);
      case 'm':		/* -mime */
	omime = 1;
	return(1);
      case 'w':		/* Width values  */
	for (p+=2; isspace(*p); p++) ;
	if (!*p) { p = *++Argv ; --Argc ; }
	p += str2int(p, &owidth) ;
	if (ispunct(*p)) p++ ;
	p += str2int(p, &indent) ;
	if (*p) {
	    fprintf(stderr, "****Invalid numbers in: %s\n%s\n",
	    text, usage) ;
	    return(-1);
	}
	if (owidth <= 0) owidth = 80 ;
	return(1);
      case 0 :		/* Option - = take Standard Input */
	init() ;
	thefile = st_name ;
	infile = stdin ;
	do_conv(0) ;
	return(1);
      bad_option:
      default:
        fprintf(stderr, "****Invalid option: %s\n%s\n", p, usage) ;
        return(-1);
    }
    return(0);		/* Ignored Option */
}

/*==================================================================
		Main Routine
 *==================================================================*/

#ifdef FUNCTION
#else
int main(int argc, char **argv, char **envp)
{
  char *p ;
  int o ;

    _of = stdout ;
    osmtrace() ;
    /* Loop on Arguments */
    Argc = argc; Argv = argv;
    while (--Argc > 0) {
        p = *++Argv ;
        if (*p == '-') {
	    o = sgml_opt(p);
	    if (o<0) exit(1);
	    continue;
	}
	/* A File */
        init() ;
	if (infile && (infile != stdin)) fclose(infile);
	thefile = p ;
        infile = fopen(p, "r") ;
        if (!infile) {
            fprintf(stderr, "****") ;
            perror(p);
            exit(1) ;
        }
        do_conv(0) ;
    }

    if (!infile) {	/* Nothing done. Operate on stdin */
        init() ;
    	infile = stdin;
    	do_conv(0) ;
    }
    exit(0);
}
#endif
