/*++++++++++++++
.IDENTIFICATION strut.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS       Table, FITS
.VERSION  1.0   15-May-1991 (anafile)
.VERSION  4.0   28-Feb-1994 New package
.VERSION  4.1   01-Oct-1996: epochs
.VERSION  4.2   10-Jun-1997: domain
.VERSION  4.3   02-Sep-1997: Added strdup2 strdup1 strdupv
.VERSION  4.4   06-Apr-1998: Added strindex
.VERSION  4.5   18-Sep-1998: Options 'F' (FITS) and 'T' (FITS with Time) in date
.VERSION  4.6   20-Apr-2000: Option in epochs -- 1 to get any word following
.VERSION  4.7   29-May-2000: Eplit in 2 parts  strut.c util.c
.VERSION  4.8   23-Mar-2006: strbrace: accept ( { [ < > ] } )
.VERSION  4.9   09-Feb-2007: Added [^...]
.VERSION  5.0   27-Jan-2008: strbracket revisited.
.VERSION  5.1   23-Jul-2008: strdup2 accept NULL strings
.VERSION  5.2   18-Sep-2008: Added strumb1
.VERSION  5.3   02-Oct-2008: str1 accepts '\x..'..
.COMMENTS       Small C String Utilities for anafile package
---------------*/

#ifndef BZERO
#define BZERO 0
#endif

#ifndef DEBUG
#define DEBUG 0
#endif

#ifdef  OSM		/* Use specific memory.c routines */
#define malloc  osmalloc
#define free    osmfree
#define realloc osmexpand
#define strdup	osmdup
#endif
#include <stdlib.h>	/* Standard Definitions malloc   */
#include <string.h>	/* Standard string manipulations */
#include <ctype.h>	/* Standard string manipulations */

#if DEBUG
#define RETURN(x) 	{ printf("%d\n", x); recursivity--;\
		for (i=0; i<recursivity;i++) putchar(' '); \
		return(x); } 					/* Debugging */
static char inbyte;
static int i, recursivity;
#else
#define RETURN(x)	return(x)
#endif

#define ____    000     /* NULL character       */
#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])

static char empty_string[] = "";

/*==================================================================
		String utilities
 *==================================================================*/

int str1(in, byte)
/*++++++++++++++++
.PURPOSE  Get the next char according to C standards
.RETURNS  Length of interpreted byte
.REMARKS  \octal or \f \b \r \n, etc...
-----------------*/
  char *in;       /* IN: String to interpret */
  char *byte;	  /* OUT: Interpreted byte */
{
  char *p; int v;

    if (*(p=in) == '\\') switch(*++p) {
	case 'a': *byte = '\07'; return(2);	/* Alarm (bell)	*/
	case 'b': *byte = '\b'; return(2);	/* Backspace	*/
	case 'f': *byte = '\f'; return(2);	/* Form-Feed	*/
	case 'n': *byte = '\n'; return(2);	/* Newline	*/
	case 'r': *byte = '\r'; return(2);	/* Carriage-Ret	*/
	case 's': *byte = ' ';  return(2);	/* Space (blank)*/
	case 't': *byte = '\t'; return(2);	/* Tab		*/
	case 'v': *byte = '\v'; return(2);	/* Vertical Tab	*/
	case 'E':				/* escape 	*/
	case 'e': *byte = 0033; return(2);
	case 'x': 
	    if (isxdigit(p[1])) {
		++p; 
		v = *p & 0xf; if (isalpha(*p)) v += 9;
		if (isxdigit(p[1])) {
		    ++p; v <<= 4;
		    v |= *p & 0xf; if (isalpha(*p)) v += 9;
		}
		*byte = v;
		return(++p - in);
	    }
	    /* NO BREAK: use default (ignore the \) */
	default: 
	    if (isdigit(*p)) {
		v = *(p++) - '0';
		if (isdigit(*p)) v = (v<<3) | (*(p++) - '0');
		if (isdigit(*p)) v = (v<<3) | (*(p++) - '0');
		*byte = v;
		return(p-in);
	    }
	    *byte = *p; return(2);
    }
    *byte = *p;
    return(*p ? 1 : 0);
}

/*==================================================================*/
int strloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;
    for ( s = text; *s; s++)      if (*s == c)    break;
    return(s-text);
}

int strbloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character backward
.RETURNS  Index of located char / -1 if not found
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;
    for (s=text+strlen(text); --s >= text; )  if (*s == c)    break;
    return(s-text);
}

/*==================================================================*/
int strmatch(template, name)
/*++++++++++++++++
.PURPOSE  Check name against template
.RETURNS  1 (Match) / 0 (mismatch)
.REMARKS  * ? [...] recognized as wild chars
-----------------*/
  char *template;	/* IN: Name template with * ? */
  char *name;		/* IN: Name to check */
{
  char *t, *f; int m, not;
#if DEBUG
    static int recursivity; int i;
    if (recursivity) { 
	putchar('\n') ; 
	for (i=0; i<recursivity; i++) putchar(' ');
    }
    recursivity++;
    printf("strmatch(\"%s\", \"%s\") => ", template, name);
#endif
    for (t = template, f = name; *t; t++) switch(*t) {
      case '?':			/* Matches one character */
	if (*f) f++;
	else RETURN(0);
	continue;
      case '*':			/* Match 0 or n characters */
	if (! *++t) RETURN(1);	/* The * is the last char of template */
#if 1		/* Try LONGEST possible matching */
	for (m = strlen(f); --m >= 0; ) {
	    if (strmatch(t, f+m))
		RETURN(1);
	}
#else		/* Try SHORTEST possible matching */
	for (; *f; f++) {	/* Look for first matching char */
	    if (isalnum(*t) && (*f != *t)) continue; 
	    if (strmatch(t,f))	/* Recursive call */
		RETURN(1);
	}
#endif
	RETURN(0);
      case '[':		/* A set of characters is provided */
	++t; m=not=0;
	if (*t == '^') { not = 1 ; t++ ; }
	if (*t == ']') { /* Closing brace is acce[table	*/
	    t++;
	    m = (*f == ']');
	}
	while ((m==0) && *t) {	/* Check: does *f be in set ? 	*/
	    if (*t == ']') RETURN(0);		/* End of set	*/
	    if ((t[1] == '-') && (t[2] != ']')) {      /* Range	*/
		m = (*f >= t[0]) && (*f <= t[2]);
		t += 3;
	    }
	    else m = (*f == *t), t++;
	}
	m ^= not;	/* V4.9 */
	if (m == 0) RETURN(0);
	t += strloc(t, ']');
	f++;
	continue;
      case '\\':	/* Escaped Char. NO BREAK HERE! */
	t++;
      default:
	if (*t != *f) RETURN(0);
	f++;
	continue;
    }
    RETURN((!*t) && (!*f));	/* Returns 1 if *t AND *f are NUL */
}

/*==================================================================*/
int stumatch(template, name)
/*++++++++++++++++
.PURPOSE  Check name against template, case insensitive
.RETURNS  1 (Match) / 0 (mismatch)
.REMARKS  Only * ? recognized as wild chars
-----------------*/
  char *template;	/* IN: Name template with * ? */
  char *name;		/* IN: Name to check */
{
  char *t, *f, c; int m, not;
    for (t = template, f = name; *t; t++) switch(*t) {
      case '?':			/* Matches one character */
	if (*f) f++;
	else RETURN(0);
	continue;
      case '*':			/* Match 0 or n characters */
	if (! *++t) RETURN(1);	/* The * is the last char of template */
#if 1		/* Try LONGEST possible matching */
	for (m = strlen(f); --m >= 0; ) {
	    if (stumatch(t, f+m))
		RETURN(1);
	}
#else		/* Try SHORTEST possible matching */
	for (; *f; f++) {	/* Look for first matching char */
	    if (isalnum(*t) && (*f != *t)) continue; 
	    if (stumatch(t,f))	/* Recursive call */
		RETURN(1);
	}
#endif
	RETURN(0);
      case '[':		/* A set of characters is provided */
	++t; m=not=0; c = toupper(*f);
	if (*t == '^') { not = 1 ; t++ ; }
	if (*t == ']') { /* Closing brace is acce[table	*/
	    t++;
	    m = (c == ']');
	}
	while ((m==0) && *t) {	/* Check: does c be in set ? 	*/
	    if (*t == ']') RETURN(0);		/* End of set	*/
	    if ((t[1] == '-') && (t[2] != ']')) {      /* Range	*/
		m = (c >= toupper(t[0])) && (c <= toupper(t[2]));
		t += 3;
	    }
	    else m = (c == toupper(*t)), t++;
	}
	m ^= not;
	if (m == 0) RETURN(0);
	t += strloc(t, ']');
	f++;
	continue;
      case '\\':	/* Escaped Char. NO BREAK HERE! */
	t++;
      default:
	if (toupper(*t) != toupper(*f)) RETURN(0);
	f++;
	continue;
    }
    return((!*t) && (!*f));	/* Returns 1 if *t AND *f are NUL */
}

/*==================================================================
		Allocating and Concatenating Strings
 *==================================================================*/

char *strdupv(argv)
/*++++++++++++++++
.PURPOSE  Create the concatenation of n strings into a newly allocated piece
	(last argument must be NULL)
.RETURNS  The concatenation
-----------------*/
  char **argv ;	/* IN: Vector od strings to concatenate */
{
  char **a, *r, *p ;
  int len ;
    for (len=1, a=argv; *a; a++) len += strlen(*a) ;
    p = r = malloc(len) ;
    for (a=argv; *a; a++) strcpy(p, *a), p += strlen(p) ;
    return(r) ;
}

char *strdup2(s1, s2)
/*++++++++++++++++
.PURPOSE  Create the concatenation of 2 strings into a newly allocated piece
.RETURNS  The concatenation
.REMARKS  Accept NULL strings
-----------------*/
  char *s1, *s2 ;
{
  char *s ;
    if (!s1) s1 = empty_string;
    if (!s2) s2 = empty_string;
    s = malloc(strlen(s1) + strlen(s2) + 1) ;
    strcpy(s, s1) ;
    strcat(s, s2) ;
    return(s) ;
}

char *strdup3(s1, s2, s3)
/*++++++++++++++++
.PURPOSE  Create the concatenation of 3 strings into a newly allocated piece
.RETURNS  The concatenation
-----------------*/
  char *s1, *s2, *s3 ;
{
  char *p, *s ; int len[2] ;
    if (!s1) s1 = empty_string;
    if (!s2) s2 = empty_string;
    if (!s3) s3 = empty_string;
    len[0] = strlen(s1) ; len[1] = strlen(s2) ;
    s = p = malloc(len[0] + len[1] + strlen(s3) + 1) ;
    strcpy(p, s1) ; p += len[0] ;
    strcpy(p, s2) ; p += len[1] ;
    strcpy(p, s3) ;
    return(s) ;
}

char *strdupn(char *s1, int len)
/*++++++++++++++++
.PURPOSE  Duplicate a part of a string
.RETURNS  The copy
-----------------*/
{
  char *s ;
    if (len < 0) len = strlen(s1) ;
    s = malloc(len + 1) ;
    strncpy(s, s1, len) ;
    s[len] = 0 ;
    return(s) ;
}

int strbracket(char *text)
/*++++++++++++++++
.PURPOSE  In a text starting by ( [ { < compute length until closing > } ] ) 
.RETURNS  Index of  closing bracket
-----------------*/
{
#if 0	/* Old code */
  char sym[2];
  int depth; char *p ;
    p = text;
    sym[0] = *(p++);
    if (sym[0] == '(') sym[1] = ')';
    else if (sym[0] == '[') sym[1] = ']';
    else if (sym[0] == '{') sym[1] = '}';
    else if (sym[0] == '<') sym[1] = '>';
    else return(0);
    for (depth = 1; depth && *p ; p++) switch(*p) {
	case '\020': 
	case '\021': 
    	case '\\': p++; continue;
	case '(': case '[': case '{': case '<':
	    if (*p == sym[0]) depth++;
	    continue;
	case ')': case ']': case '}': case '>':
	    if (*p == sym[1]) depth--;
	    continue;
    }
    if (depth == 0) p-- ;
    return(p-text);
#else
static char brackets[] = "()[]{}<>";
  int depth=1, ip; char *p ;
    p = text;
    ip = strloc(brackets, *text);
    if (brackets[ip])           /* Text does start by opening bracket */
        p++;
    else                        /* Text does not start by bracket ... */
        return(0);
    while (depth && *p) {
        if ((*p == '\020') || (*p == '\021') || (*p == '\\')) ;
        else if (*p == brackets[ip])     depth++;
        else if (*p == (brackets+1)[ip]) depth--;
        p++;
    }
    if (depth == 0) p-- ;
    return(p-text);
#endif
}

/*==================================================================
		Unicode
 *==================================================================*/
int strumb1(char *text)
/*++++++++++++++++
.PURPOSE  Convert some Unicode multi-byte extensions
.RETURNS  Number of substitutions
.REMARKS  text is MODIFIED!
    Conversions include: nbsp --> ' '
    minus [mn]sahs variations --> '-'
-----------------*/
{
  char *p, *t;
  int n=0;
    p = t = text;
    while (*t) {
	if ((*t)&0x80) {      /* Extension */
	    if (*t == '\xa0') /* nbsp */               { *t = ' '; n++; }
	    else if ((*t == '\x96') || (*t == '\x97')) { *t = '-'; n++; }
	    else if (*t == '\xe2') {	/* Multi-byte math. extension */
		if (((t[1] == '\x88') && (t[2] == '\x92')) ||
		    ((t[1] == '\x80') &&((t[2]&0xfc)==0x90))) {
		    *(p++) = '-'; 
		    t += 3;
		    n++;
		    continue;
		}
		if (t[1] && t[2]) {
		    *(p++) = *(t++);
		    *(p++) = *(t++);
		}
	    }
	}
	if (n) *p = *t;
	p++; t++;
    }
    if (n) *p = 0;
    return(n);
}
